# MIT License

# Copyright (c) 2024 Lars Andersen Bratholm

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# https://github.com/optuna/optuna/issues/5299

"""
Variation of the multivariate group-decomposed TPESampler, that samples parameter subspaces in a hierarchical
fashion. E.g. if the possible parameter combinations that can be sampled are {x, y, a} and {x, y, b}, then the
parameter subspaces are.

[{x, y}, {a}, {b}]. Since x and y are always sampled, n_ei_candidates sampled of {x, y} is generated. Then a
classifier estimates which of the sampled parameter values will result in 'a' being sampled, and which will
result in 'b' being sampled. Then either 'a' or 'b' is sampled depending on the values of {x, y}.
"""

import functools
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import optuna
import sklearn.preprocessing
from loguru import logger
from numpy.typing import NDArray
from optuna.distributions import BaseDistribution, CategoricalDistribution
from optuna.samplers import TPESampler
from optuna.samplers._tpe.parzen_estimator import _ParzenEstimator
from optuna.samplers._tpe.sampler import _calculate_weights_below_for_multi_objective
from optuna.search_space import intersection_search_space
from optuna.study import Study
from optuna.trial import FrozenTrial, TrialState
from sklearn.tree import DecisionTreeClassifier
from typing_extensions import Unpack


class DefaultGroupClassifier:
    """
    Create a Decision Tree classifier on the parent group to determine which child group will be sampled.
    """

    def __init__(self) -> None:
        self.classifiers: Dict[int, DecisionTreeClassifier] = {}
        self.encoders: Dict[
            str,
            Callable[
                [NDArray[np.float64]], Union[NDArray[np.float64], NDArray[np.int_]]
            ],
        ] = {}
        self._groups: List[
            Dict[
                str,
                BaseDistribution,
            ]
        ] = []

    def __call__(
        self,
        samples: Dict[str, NDArray[np.float64]],
        trials: List[FrozenTrial],
        groups: List[
            Dict[
                str,
                BaseDistribution,
            ]
        ],
        child_group_indices: List[int],
        parent_group_index: int,
    ) -> NDArray[np.int_]:
        """
        Make predictions based on the sampled parent parameters of which child group will be sampled.
        """
        self._validate_cache(groups)
        parent_group = groups[parent_group_index]
        self._create_encoders(parent_group)
        trial_features = self._create_trial_features(trials, parent_group)
        trial_targets = self._create_trial_targets(trials, child_group_indices)
        self._fit_classifier(trial_features, trial_targets, parent_group_index)
        sample_features = self._create_sample_features(samples, parent_group)
        class_label_predictions = self.classifiers[parent_group_index].predict(
            sample_features
        )
        predictions: NDArray[np.int_] = self.classifiers[parent_group_index].classes_[
            class_label_predictions
        ]
        return predictions

    def _validate_cache(
        self,
        groups: List[
            Dict[
                str,
                BaseDistribution,
            ]
        ],
    ) -> None:
        """
        Check if.
        """
        if self._groups != groups:
            if len(self._groups) > 0:
                logger.debug(f"groups were {self._groups}, now {groups}")
            self._groups = groups
            self.classifiers = {}

    def _create_encoders(self, group: Dict[str, BaseDistribution]) -> None:
        for param_name, distribution in group.items():
            if param_name in self.encoders:
                continue
            if isinstance(distribution, CategoricalDistribution):
                choices = [
                    distribution.to_internal_repr(choice)
                    for choice in distribution.choices
                ]
                self.encoders[param_name] = (
                    lambda x: sklearn.preprocessing.label_binarize(x, classes=choices)
                )
            else:
                self.encoders[param_name] = lambda x: np.asarray(x).reshape(-1, 1)

    def _create_trial_features(
        self, trials: List[FrozenTrial], group: Dict[str, BaseDistribution]
    ) -> NDArray[np.float64]:
        features = []
        for param_name, distribution in group.items():
            internal_repr: NDArray[np.float64] = np.asarray(
                [
                    distribution.to_internal_repr(trial.params[param_name])
                    for trial in trials
                ]
            )
            param_features = self.encoders[param_name](internal_repr)
            features.append(param_features)
        features_: NDArray[np.float64] = np.concatenate(features, axis=1)
        assert len(trials) == features_.shape[0]
        return features_

    def _create_trial_targets(
        self, trials: List[FrozenTrial], group_indices: List[int]
    ) -> NDArray[np.int_]:
        n_groups = len(group_indices)
        targets = []
        for trial in trials:
            trial_params = set(trial.params)
            for target_idx, group_idx in enumerate(group_indices):
                if trial_params.issuperset(self._groups[group_idx]):
                    targets.append(target_idx)
                    break
            else:
                # No children in trial
                targets.append(n_groups)

        targets_ = np.asarray(targets)
        assert targets_.shape[0] == len(trials)
        return targets_

    def _fit_classifier(
        self, features: NDArray[np.float64], targets: NDArray[np.int_], index: int
    ) -> None:
        # Use cached classifier if it correctly classifies
        # all trials
        if index in self.classifiers:
            predictions = self.classifiers[index].predict(features)
            if (predictions == targets).all():
                return

        n_samples = features.shape[0]
        # Keep increasing depth until all samples are classified correctly
        for depth in range(1, n_samples):
            classifier = DecisionTreeClassifier(max_depth=depth, max_features=1)
            classifier.fit(features, targets)
            predictions = classifier.predict(features)
            if (predictions == targets).all():
                self.classifiers[index] = classifier
                return

        raise RuntimeError("Could not fit classifier")

    def _create_sample_features(
        self,
        samples: Dict[str, NDArray[np.float64]],
        parent_group: Dict[str, BaseDistribution],
    ) -> NDArray[np.float64]:
        sample_features: List[Union[NDArray[np.float64], NDArray[np.int_]]] = []
        for param_name in parent_group:
            param_features: Union[NDArray[np.float64], NDArray[np.int_]] = (
                self.encoders[param_name](samples[param_name])
            )
            sample_features.append(param_features)
        sample_features_: NDArray[np.float64] = np.concatenate(sample_features, axis=1)
        return sample_features_


def objective_classifier(
    samples: Dict[str, NDArray[np.float64]],
    trials: List[FrozenTrial],
    groups: List[Dict[str, BaseDistribution]],
    child_group_indices: List[int],
    parent_group_index: int,
) -> NDArray[np.int_]:
    """
    Manual classifier for the benchmark objective.
    """
    parent_group = groups[parent_group_index]

    if set(parent_group) == set(("x", "y")):
        predictions = np.zeros(samples["x"].shape, dtype=int)
        samples_external_repr = np.asarray(
            [parent_group["x"].to_external_repr(sample) for sample in samples["x"]],
            dtype=bool,
        )
        assert len(child_group_indices) == 2
        for i, idx in enumerate(child_group_indices):
            child_group = set(groups[idx])
            if "n" in child_group:
                predictions[samples_external_repr == True] = i
            else:
                assert "m" in child_group
                predictions[samples_external_repr == False] = i
    elif set(parent_group) == set(("n",)):
        predictions = np.zeros(samples["n"].shape, dtype=int)
        samples_external_repr = np.asarray(
            [parent_group["n"].to_external_repr(sample) for sample in samples["n"]],
            dtype=bool,
        )
        for i, idx in enumerate(child_group_indices):
            child_group = set(groups[idx])
            if "a" in child_group:
                predictions[samples_external_repr == True] = i
            else:
                assert "b" in child_group
                predictions[samples_external_repr == False] = i
    else:
        assert set(parent_group) == set(("m",))
        predictions = np.zeros(samples["m"].shape, dtype=int)
        samples_external_repr = np.asarray(
            [parent_group["m"].to_external_repr(sample) for sample in samples["m"]],
            dtype=bool,
        )
        for i, idx in enumerate(child_group_indices):
            child_group = set(groups[idx])
            if "c" in child_group:
                predictions[samples_external_repr == True] = i
            else:
                assert "d" in child_group
                predictions[samples_external_repr == False] = i
    return predictions


class HierarchicalTPESampler(TPESampler):
    """
    The group-decomposed TPESampler, samples each group independently.

    Often parameters in each group can still be correlated.
    This sampler utilizes that the intersection search space
    of the trials containing the parameters of a group can be
    the joint set of both the group itself and other groups as well.
    Each observed group combination is sampled from and an
    additional term is added to the likelihood representing the
    population of each group in below/above.
    """

    def __init__(
        self,
        *args: Unpack[Sequence[Any]],  # type: ignore
        classifier: Optional[
            Union[
                Callable[
                    [
                        Dict[str, NDArray[np.float64]],
                        List[FrozenTrial],
                        List[Dict[str, BaseDistribution]],
                        List[int],
                        int,
                    ],
                    NDArray[np.int_],
                ],
                DefaultGroupClassifier,
            ]
        ] = None,
        **kwargs: Unpack[Dict[str, Any]],  # type: ignore
    ) -> None:
        if classifier is None:
            self._classifier: Union[
                Callable[
                    [
                        Dict[str, NDArray[np.float64]],
                        List[FrozenTrial],
                        List[Dict[str, BaseDistribution]],
                        List[int],
                        int,
                    ],
                    NDArray[np.int_],
                ],
                DefaultGroupClassifier,
            ] = DefaultGroupClassifier()
        else:
            self._classifier = classifier
        self._tpe_sampler = TPESampler(*args, multivariate=False, group=False, **kwargs)

        super().__init__(*args, multivariate=True, group=True, **kwargs)

    def sample_relative(
        self,
        study: Study,
        trial: FrozenTrial,
        search_space: Dict[str, BaseDistribution],
    ) -> Dict[str, Any]:
        assert self._search_space_group is not None

        if search_space == {}:
            return {}

        states = (TrialState.COMPLETE, TrialState.PRUNED)
        trials = study._get_trials(deepcopy=False, states=states, use_cache=True)
        # If the number of samples is insufficient, we run random trial.
        if len(trials) < self._n_startup_trials:
            return {}

        params = self._sample_hierarchy(study, search_space)

        # The unsampled parameters will likely not be used during evaluation,
        # but can be if the trained classifier makes an incorrect prediction
        for param_name, param_distribution in search_space.items():
            if param_name not in params:
                params[param_name] = self._tpe_sampler.sample_independent(
                    study, trial, param_name, param_distribution
                )
        return params

    def _get_trials(self, study: Study) -> List[FrozenTrial]:
        """
        Get complete, pruned and possibly running trials.
        """
        if self._constant_liar:
            states = [TrialState.COMPLETE, TrialState.PRUNED, TrialState.RUNNING]
        else:
            states = [TrialState.COMPLETE, TrialState.PRUNED]
        use_cache = not self._constant_liar
        trials = study._get_trials(deepcopy=False, states=states, use_cache=use_cache)
        return trials

    def _get_group_probabilities(
        self, group_weights: List[List[List[float]]]
    ) -> Tuple[NDArray[np.float64], NDArray[np.float64]]:
        """
        Get the group probabilities with prior (symmetrical dirichlet with alpha=1) if required.
        """
        n_groups = len(group_weights[0])
        below_weights = np.asarray(
            functools.reduce(lambda x, y: x + y, group_weights[0])
        )
        above_weights = np.asarray(
            functools.reduce(lambda x, y: x + y, group_weights[1])
        )
        p_group_below = np.zeros(n_groups)
        p_group_above = np.zeros(n_groups)
        assert self._parzen_estimator_parameters.prior_weight is not None
        if self._parzen_estimator_parameters.consider_prior is True:
            n_effective_samples_below = (
                HierarchicalTPESampler._get_effective_sample_size(below_weights)
            )
            n_effective_samples_above = (
                HierarchicalTPESampler._get_effective_sample_size(above_weights)
            )
            for i, weights in enumerate(group_weights[0]):
                p_group_below[i] = (
                    sum(weights) / sum(below_weights) * n_effective_samples_below
                    + self._parzen_estimator_parameters.prior_weight * 1.0
                )
            for i, weights in enumerate(group_weights[1]):
                p_group_above[i] = (
                    sum(weights) / sum(above_weights) * n_effective_samples_above
                    + self._parzen_estimator_parameters.prior_weight * 1.0
                )
        else:
            for i, weights in enumerate(group_weights[0]):
                p_group_below[i] = sum(weights) / sum(below_weights)
            for i, weights in enumerate(group_weights[1]):
                p_group_above[i] = sum(weights) / sum(above_weights)

        # Normalize
        p_group_below /= p_group_below.sum()
        p_group_above /= p_group_above.sum()
        return p_group_below, p_group_above

    def _get_group_trials_and_weights(
        self,
        study: Study,
        trials: List[FrozenTrial],
        search_spaces: List[Dict[str, BaseDistribution]],
        parent_search_space: Optional[Dict[str, BaseDistribution]] = None,
    ) -> Tuple[List[List[List[FrozenTrial]]], List[List[List[float]]]]:
        """
        Get the trial subsets and group weights for the given search spaces.

        :param study: the study
        :param trials: observed trials
        :param search_spaces: the overlapping group search spaces
        :returns: group trials and group weights for below and above
        """
        # Split trials into below/above and get their weights
        below_trials, above_trials, weights_below, weights_above = (
            self._split_trials_weights(trials, study)
        )

        # Gather the trials belonging to each group, with
        # index=0 for below and index=1 for above.
        group_trials: List[List[List[FrozenTrial]]] = [
            [[] for _ in search_spaces],
            [[] for _ in search_spaces],
        ]
        group_weights: List[List[List[float]]] = [
            [[] for _ in search_spaces],
            [[] for _ in search_spaces],
        ]
        if parent_search_space is not None:
            group_trials[0].append([])
            group_trials[1].append([])
            group_weights[0].append([])
            group_weights[1].append([])
        # First assign trials where the intersection search space match a search space exactly
        intersection = intersection_search_space(trials)
        intersection_search_space_index = None
        for k, search_space in enumerate(search_spaces):
            if set(intersection) == set(search_space):
                intersection_search_space_index = k
        n_below = len(below_trials)
        for i, (trial, weight) in enumerate(
            zip(
                below_trials + above_trials,
                weights_below.tolist() + weights_above.tolist(),
            )
        ):
            below_above_index = int(i < n_below)
            # First assign trials where the intersection search space match a search space exactly
            if intersection_search_space_index is not None:
                if set(trial.params) == set(
                    search_spaces[intersection_search_space_index]
                ):
                    group_trials[below_above_index][
                        intersection_search_space_index
                    ].append(trial)
                    group_weights[below_above_index][
                        intersection_search_space_index
                    ].append(weight)
                    assert False
                    continue
            if parent_search_space is not None and set(trial.params) == set(
                parent_search_space
            ):
                group_trials[below_above_index][-1].append(trial)
                group_weights[below_above_index][-1].append(weight)
                assert False
                continue

            n_assignments = 0
            # Then relax to supersets excluding the intersection search space group
            for k, search_space in enumerate(search_spaces):
                if k == intersection_search_space_index:
                    continue
                if set(trial.params).issuperset(search_space):
                    n_assignments += 1
                    group_trials[below_above_index][k].append(trial)
                    group_weights[below_above_index][k].append(weight)
            assert n_assignments < 2
            if parent_search_space is not None:
                assert set(trial.params).issuperset(parent_search_space)
                if n_assignments == 0:
                    group_trials[below_above_index][-1].append(trial)
                    group_weights[below_above_index][-1].append(weight)

        # Check that all trials are assigned
        assert len(
            functools.reduce(
                lambda x, y: x + y,
                functools.reduce(lambda x, y: x + y, group_trials),  # type: ignore
            )
        ) == len(trials)
        return group_trials, group_weights

    def _split_trials_weights(
        self, trials: List[FrozenTrial], study: Study
    ) -> Tuple[
        List[FrozenTrial], List[FrozenTrial], NDArray[np.float64], NDArray[np.float64]
    ]:
        # Divide data into below and above.
        n_trials = sum(
            trial.state != TrialState.RUNNING for trial in trials
        )  # Ignore running trials.
        below_trials, above_trials = optuna.samplers._tpe.sampler._split_trials(
            study,
            trials,
            self._gamma(n_trials),
            self._constraints_func is not None,
        )
        # Calculate weights of trials
        if study._is_multi_objective():
            weights_below = _calculate_weights_below_for_multi_objective(
                study, below_trials, self._constraints_func
            )
        else:
            weights_below = self._parzen_estimator_parameters.weights(len(below_trials))
            assert len(below_trials) == len(weights_below)
        weights_above = self._parzen_estimator_parameters.weights(len(above_trials))

        # Normalize
        weights_below /= weights_below.sum()
        weights_above /= weights_above.sum()
        return below_trials, above_trials, weights_below, weights_above

    def _get_mpes(
        self,
        study: Study,
        trials: List[FrozenTrial],
        search_space: Dict[str, BaseDistribution],
    ) -> Tuple[_ParzenEstimator, _ParzenEstimator]:
        """
        Generate samples with the Parzen estimator, using the trial subset containing all parameters in the
        given search space.

        :param study: the study containing observed trials
        :param trials: the trials with the same intersection search space as the given search space
        :param search_space: the search space to expore
        :returns: the samples and the log-likelihood difference between below and above
        """
        assert all(set(trial.params).issuperset(search_space) for trial in trials)

        # We divide data into below and above.
        n = sum(
            trial.state != TrialState.RUNNING for trial in trials
        )  # Ignore running trials.
        below_trials, above_trials = optuna.samplers._tpe.sampler._split_trials(
            study,
            trials,
            self._gamma(n),
            self._constraints_func is not None,
        )

        below = self._get_internal_repr(below_trials, search_space)
        above = self._get_internal_repr(above_trials, search_space)

        # We then sample by maximizing log likelihood ratio.
        if study._is_multi_objective():
            weights_below = _calculate_weights_below_for_multi_objective(
                study, below_trials, self._constraints_func
            )
            mpe_below = _ParzenEstimator(
                below, search_space, self._parzen_estimator_parameters, weights_below
            )
        else:
            mpe_below = _ParzenEstimator(
                below, search_space, self._parzen_estimator_parameters
            )
        mpe_above = _ParzenEstimator(
            above, search_space, self._parzen_estimator_parameters
        )

        return mpe_below, mpe_above

    def _sample_hierarchy(
        self,
        study: Study,
        search_space: Dict[str, Any],
    ) -> Dict[str, Any]:
        """
        Generate samples and select the one with the highest likelihood ratio between below and above.

        :param study: the study containing observed trials
        :param search_space: the full observed search space
        """
        # Get trials
        trials = self._get_trials(study)

        # Determine the hierachy from observed trials
        hierarchy = self._determine_hierarchy(trials)
        assert self._search_space_group is not None
        groups = self._search_space_group.search_spaces

        # Generate samples hierarchically and get the log likelihood difference
        # between below and above
        mask, samples, delta_ll = self._get_hierarchical_samples(
            study, hierarchy, groups, trials, self._n_ei_candidates
        )

        # Select the best overall sample
        ret = {}
        best_sample_index = np.argmax(delta_ll)
        mask_to_idx = np.empty(self._n_ei_candidates, dtype=int)
        for param_name, distribution in search_space.items():
            if mask[param_name][best_sample_index] is np.bool_(True):
                mask_to_idx[mask[param_name]] = np.arange(mask[param_name].sum())
                idx = mask_to_idx[best_sample_index]
                ret[param_name] = distribution.to_external_repr(
                    samples[param_name][idx]
                )
        return ret

    def _get_samples(
        self,
        study: Study,
        trials: List[FrozenTrial],
        group: Dict[str, BaseDistribution],
        n_samples: int,
    ) -> Tuple[Dict[str, NDArray[np.float64]], NDArray[np.float64]]:
        trial_subset = [trial for trial in trials if set(group).issubset(trial.params)]
        mpe_below, mpe_above = self._get_mpes(study, trial_subset, group)
        samples = mpe_below.sample(self._rng.rng, n_samples)
        assert set(samples) == set(group)
        log_likelihoods_below = mpe_below.log_pdf(samples)
        log_likelihoods_above = mpe_above.log_pdf(samples)
        delta_ll = log_likelihoods_below - log_likelihoods_above
        return samples, delta_ll

    def _get_hierarchical_samples(
        self,
        study: Study,
        hierarchy: List[Optional[int]],
        groups: List[Dict[str, BaseDistribution]],
        trials: List[FrozenTrial],
        n_samples: int,
    ) -> Tuple[
        Dict[str, NDArray[np.bool_]],
        Dict[str, NDArray[np.float64]],
        NDArray[np.float64],
    ]:
        samples = {}
        mask = {}
        for group in groups:
            for param_name in group:
                mask[param_name] = np.zeros(n_samples, dtype=bool)
        delta_ll = np.zeros(n_samples)
        n_groups = len(hierarchy)
        # Sample the current top-level groups
        for i in range(n_groups):
            if hierarchy[i] is not None:
                continue
            group_samples, d_ll = self._get_samples(
                study, trials, groups[i], n_samples=n_samples
            )
            for param_name in group_samples:
                assert param_name not in samples
                # if param_name not in samples:
                mask[param_name][:] = True
                samples[param_name] = group_samples[param_name]
            delta_ll += d_ll

            self._get_hierarchical_child_samples(
                study,
                hierarchy,
                groups,
                trials,
                mask,  # type: ignore
                samples,
                delta_ll,
                parent_group_index=i,
            )

            assert all(mask[param_name].all() for param_name in groups[i])
        return mask, samples, delta_ll  # type: ignore

    def _get_hierarchical_child_samples(
        self,
        study: Study,
        hierarchy: List[Optional[int]],
        groups: List[Dict[str, BaseDistribution]],
        trials: List[FrozenTrial],
        mask: Dict[str, NDArray[np.bool_]],
        samples: Dict[str, NDArray[np.float64]],
        delta_ll: NDArray[np.float64],
        parent_group_index: int,
    ) -> None:
        """
        Generate samples for each group having the given group as parent.
        """
        n_groups = len(hierarchy)
        child_group_indices = [
            index for index in range(n_groups) if hierarchy[index] == parent_group_index
        ]
        n_child_groups = len(child_group_indices)
        if n_child_groups == 0:
            return

        parent_group = groups[parent_group_index]
        trials = [
            trial for trial in trials if set(trial.params).issuperset(parent_group)
        ]
        predictions = self._classifier(
            samples, trials, groups, child_group_indices, parent_group_index
        )

        # Get the subsets of below and above trials belonging to each group, and the relative weights
        # of each group
        _, group_weights = self._get_group_trials_and_weights(
            study,
            trials,
            [groups[index] for index in child_group_indices],
            parent_group,
        )
        # Get group counts
        p_group_below, p_group_above = self._get_group_probabilities(group_weights)

        for i, index in enumerate(child_group_indices):
            prediction_mask = predictions == i
            if sum(prediction_mask) == 0:
                continue
            child_hierarchy = [
                -1 if item is None else None if j == index else item
                for j, item in enumerate(hierarchy)
            ]
            child_mask, child_samples, child_delta_ll = self._get_hierarchical_samples(
                study=study,
                hierarchy=child_hierarchy,
                groups=groups,
                trials=trials,
                n_samples=sum(prediction_mask),
            )
            # Add in delta log likelihoods for the group contribution
            delta_ll[prediction_mask] += (
                child_delta_ll + np.log(p_group_below[i]) - np.log(p_group_above[i])
            )
            for param_name in child_samples:
                assert param_name not in samples
                mask[param_name][prediction_mask] = child_mask[param_name]
                samples[param_name] = child_samples[param_name]

        # Add delta ll for when there is no children groups
        prediction_mask = predictions == n_groups
        if sum(prediction_mask) > 0:
            delta_ll[prediction_mask] += np.log(p_group_below[n_groups]) - np.log(
                p_group_above[n_groups]
            )

    def _determine_hierarchy(self, trials: List[FrozenTrial]) -> List[Optional[int]]:
        """
        :param trials: the observed trials
        """
        trials = [trial for trial in trials if trial.state != TrialState.RUNNING]
        assert self._search_space_group is not None
        groups = self._search_space_group.search_spaces
        n_groups = len(groups)
        hierarchy: List[List[Optional[int]]] = [[] for _ in range(n_groups)]
        for j, child_group in enumerate(groups):
            child_group_intersection_search_space = intersection_search_space(
                [trial for trial in trials if set(trial.params).issuperset(child_group)]
            )
            for i, parent_group in enumerate(groups):
                if i == j:
                    continue
                if set(parent_group).issubset(child_group_intersection_search_space):
                    hierarchy[j].append(i)
        # Clean so each group only keeps the direct parent
        finished = False
        counter = 0
        while not finished:
            counter += 1
            if counter >= 1000:
                raise RuntimeError(
                    "Unable to determine a strict hierarchy between parameter groups"
                )
            finished = True
            for item in hierarchy:
                if len(item) <= 1:
                    continue
                finished = False
                for index1 in item:
                    for index2 in item:
                        if index1 == index2:
                            continue
                        assert index2 is not None
                        if index1 in hierarchy[index2]:
                            item.pop(item.index(index1))
                            break
                    else:
                        continue
                    break
                break
        # Finally change the format to make it easier to work with
        hierarchy_ = [item[0] if len(item) == 1 else None for item in hierarchy]
        return hierarchy_

    @staticmethod
    def _get_effective_sample_size(weights: NDArray[np.float64]) -> float:
        """
        Get the effective sample size given a set of weights.

        :param weights: the weights
        :returns: the effective sample size
        """
        effective_sample_size = float(weights.sum() ** 2 / (weights**2).sum())
        return effective_sample_size



import nltk

from syftr.configuration import cfg
from syftr.utils.locks import distributed_lock


def download_nltk_data():
    with distributed_lock("nltk_download", host_only=True):
        nltk.download("punkt", cfg.paths.nltk_dir, quiet=True)
        nltk.download("stopwords", cfg.paths.nltk_dir, quiet=True)


def prepare_worker():
    download_nltk_data()


if __name__ == "__main__":
    prepare_worker()



import pytest

from syftr.evaluation.evaluator_factory import json_parser_function


@pytest.mark.parametrize(
    "data,expected_score,expected_reasoning",
    [
        (
            '{"score": 1e2, "reasoning": "Looks good."}',
            1e2,
            "Looks good.",
        ),
        (
            'Reasoning missing {"score": 1e2}',
            None,
            None,
        ),
        (
            'Reasoning empty {"score": 1e2, "reasoning": ""}',
            None,
            None,
        ),
        (
            'Bad score {"score": "top", "reasoning": "best I have seen so far"}',
            None,
            None,
        ),
        (
            'Score missing {"reasoning": "best I have seen so far"}',
            None,
            None,
        ),
        (
            'With additional text {"score": 0.8, "reasoning": "Looks good."} Some text after',
            0.8,
            "Looks good.",
        ),
        (
            'Two JSONs {"score": 0.5, "reasoning": "First."} Second {"score": 0.9, "reasoning": "Second."}',
            0.9,
            "Second.",
        ),
        (
            """Proper nesting {"score": 0.5, "reasoning": {"score": 0.9, "reasoning": "Second."}} Second""",
            None,
            None,
        ),
        ("No JSON here!", None, None),
        ('Invalid JSON {"score": 0.7, "reasoning": "Oops",}', None, None),
        ('Bad keys {"foo": 1, "bar": 2}', None, None),
        (
            'With string score {"score": "1.0", "reasoning": "String score"}',
            1.0,
            "String score",
        ),
        (
            """Nested and normal {"score": 0.5, "reasoning": {"score": 0.9, "reasoning": "Second."}} Some text {"score": 0.8, "reasoning": "Looks good."} Some text after""",
            0.8,
            "Looks good.",
        ),
        (
            """Bad formatting and newlines
{

  "score": 1.0,

   "reasoning": "with bad formatting and newlines"

}""",
            1.0,
            "with bad formatting and newlines",
        ),
        (
            '''Nesting and formatting 
{
  "score": 0.5, 
  "reasoning": """{"score": 0.9, "reasoning": "Second."}""""
} 
Some text 
{
  "score": 0.8, 
  "reasoning": "Looks good."
}
Some text after''',
            0.8,
            "Looks good.",
        ),
        (
            """Last one nested
{
  "score": 0.8, 
  "reasoning": "Looks good."
}
{
  "score": 0.5, 
  "reasoning": {"score": 0.5, "reasoning": "all fine?"}
} """,
            None,
            None,
        ),
    ],
)
def test_json_parser(data, expected_score, expected_reasoning):
    score, reasoning = json_parser_function(data)
    assert score == expected_score
    assert reasoning == expected_reasoning



import typing as T
from pathlib import Path

import yaml

from syftr.ray.submit import get_client, start_study, tail
from syftr.storage import JudgeEvalHF
from syftr.studies import (
    ConsensusCorrectnessEvaluator,
    Evaluation,
    JudgeSearchSpace,
    OptimizationConfig,
    RandomCorrectnessEvaluator,
    SingleCorrectnessEvaluator,
    StudyConfig,
    get_llm_name_combinations,
)

JUDGE_LLMS: T.List[str] = [
    "gpt-4o-mini",
    "qwen-235b-a22b",
    "glm-4.5-air",
    "gpt-oss-120b-low",
    "gpt-oss-120b-medium",
    "gpt-oss-120b-high",
    "gpt-oss-20b-low",
    "gpt-oss-20b-medium",
    "gpt-oss-20b-high",
    "nemotron-super-49b",
    "qwen3-30b-a3b",
    "gemma3-27b-it",
    "phi-4-multimodal-instruct",
]


def main():
    # name = "judge-eval-consensus"
    name = "judge-eval-study-17"
    study_config = StudyConfig(
        name=name,
        reuse_study=False,
        recreate_study=True,
        dataset=JudgeEvalHF(),
        evaluation=Evaluation(mode="judge"),
        search_space=JudgeSearchSpace(
            judge_prompts=[
                "default",
                "simple",
                "out_of_ten",
                "detailed",
            ],
            single_correctness_evaluator=SingleCorrectnessEvaluator(
                response_synthesizer_llm_names=JUDGE_LLMS
            ),
            consensus_correctness_evaluator=ConsensusCorrectnessEvaluator(
                response_synthesizer_llm_combinations=get_llm_name_combinations(
                    JUDGE_LLMS, [3]
                ),
            ),
            random_correctness_evaluator=RandomCorrectnessEvaluator(
                response_synthesizer_llm_combinations=get_llm_name_combinations(
                    JUDGE_LLMS, [3]
                ),
            ),
        ),
        optimization=OptimizationConfig(
            num_trials=2000,
            baselines=[],
            num_random_trials=50,
            use_individual_baselines=False,
            use_agent_baselines=False,
            use_variations_of_baselines=False,
            max_concurrent_trials=50,
            num_eval_samples=1000,
            num_eval_batch=10,
            max_eval_failure_rate=0.05,
        ),
    )
    path = Path("studies") / f"{name}.yaml"
    with open(path, "w") as f:
        yaml.dump(study_config.dict(), f)

    client = get_client()
    job_id = start_study(
        client=client,
        study_config_file=path,
        study_config=study_config,
        delete_confirmed=True,
    )

    tail(client, job_id)


if __name__ == "__main__":
    main()



import json
import typing as T

from llama_index.core.prompts import RichPromptTemplate

from syftr.configuration import cfg
from syftr.logger import logger

# Default prompt components
_DEFAULT_INSTRUCTIONS = """
    You are a helpful assistant.
    Answer the provided question given the context information and not prior knowledge.
"""
_CONTEXT = """
    Context information is below.
    ---------------------{context_str}---------------------
"""
_QUERY_STR = """
    Question: {query_str}
    Answer:
"""

# Concise prompt components.
_CONCISE_INSTRUCTIONS = """
    You are a helpful assistant.
    Answer the provided question given the context information and not prior knowledge.
    Be concise!
"""

# CoT prompt components.
_COT_INSTRUCTIONS = """
    You are a helpful assistant.
    Answer the provided question given the context step-by-step.
"""

# Finance expert
_FINANCE_EXPERT_INSTRUCTIONS = """
    You are an expert financial advisor, providing guidance on financial analysis and managing money.
    You need to answer the provided question given the context information and not prior knowledge.
    You are also provided with the additional knowledge you can use for performing the analysis:

    FINANCIAL SYNONYMS
    ==================
    - "(Consolidated) Balance Sheet(s)",
      "Statement(s) of (Consolidated) Financial Position", "(Consolidated) Statement(s) of Financial Position"

    - "(Consolidated) Cash Flow(s) Statement(s)", "(Consolidated) Statement(s) of Cash Flows"

    - "(Consolidated) Income Statement(s)", "Statement(s) of (Consolidated) Income", "(Consolidated) Statement(s) of Income",
      "(Consolidated) Profit-and-Loss Statement(s)", "(Consolidated) P&L (Statement(s))",
      "(Consolidated) Earnings Statement(s)", "Statement(s) of (Consolidated) Earnings", "(Consolidated) Statement(s) of Earnings",
      "(Consolidated) Operations Statement(s)", "Statement(s) of (Consolidated) Operations", "(Consolidated) Statement(s) of Operations"

    Balance-Sheet Line-Item Synonyms
    --------------------------------

    - "Total Assets", "TA(s)"

    - "(Net) Fixed Assets", "(Net) FA(s)",
      "(Net) Property, Plant & Equipment", "(Net) PP&E", "(Net) PPNE",
      "(Net) Property & Equipment", "(Net) Plant & Equipment", "(Net) Property, Equipment & Intangibles"

    - "(Total) (Net) Inventory", "(Total) (Net) Inventories",
      "(Total) (Net) Merchandise Inventory", "(Total) (Net) Merchandise Inventories"

    - "(Net) Accounts Receivable", "(Net) AR", "(Net) (Trade) Receivables"

    - "(Net) Accounts Payable", "(Net) AP"


    Cash-Flow-Statement Line-Item Synonyms
    --------------------------------------

    - "(Net) Cash (Flows) from Operations", "(Net) Cash (Flows) from Operating Activities", "(Net) Operating Cash Flows"

    - "(Net) Cash (Flows) from Investments", "(Net) Cash (Flows) from Investing Activities", "(Net) Investing Cash Flows"

    - "Capital Expenditure(s)", "CapEx", "Capital Spending", "Property, Plant & Equipment (PP&E) Expenditure(s)/Purchase(s)"

    - "(Net) Cash (Flows) from Financing", "(Net) Cash (Flows) from Financing Activities", "(Net) Financing Cash Flows"


    Income-Statement / Profit-and-Loss- (P&L-) Statement / Earnings-Statement / Operations-Statement Line-Item Synonyms
    -------------------------------------------------------------------------------------------------------------------

    - "(Total) (Net) (Operating) Revenue(s)", "(Total) (Net) Sales"

    - "(Total) Cost of Goods Sold", "(Total) COGS", "(Total) Cost of Sales", "(Total) Cost of Revenue"

    - "Gross Income (or Loss)", "Gross Profit (or Loss)", "Gross Earnings (or Loss(es))"

    - "Operating Expenditure(s)", "Operating Expenses", "OpEx"

    - "Selling, General & Administrative (Expenses)", "SG&A (Expenses)"

    - "(Unadjusted) Operating Income", "(Unadjusted) Operating Profit"

    - "Earnings before Interest, Tax, Depreciation & Amortization", "EBITDA"

    - "Depreciation & Amortization", "D&A",
      "Depreciation & Amortization of Fixed Assets", "Depreciation & Amortization of Property, Plant & Equipment (PP&E)"

    - "Earnings before Interest & Tax", "EBIT"

    - "Net Income (Attributable to Shareholders)", "Net Profit (Attributable to Shareholders)"


    FINANCIAL METRIC FORMULAS
    =========================

    Turnover Ratio Metric Formulas
    ------------------------------

    `(Total) Asset Turnover Ratio` = (
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales` /
      `average Total Assets, typically between two consecutive fiscal year-ends`
    )

    `Fixed Asset Turnover Ratio` = (
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales` /
      `average (Net) Fixed Assets, a.k.a. (Net) Property, Plant & Equipment (PP&E), typically between two consecutive fiscal year-ends`
    )

    `Inventory Turnover Ratio, a.k.a. Inventory Conversion Ratio` = (
      `(Total) Cost of Goods Sold, a.k.a. (Total) COGS, or (Total) Cost of Sales, or (Total) Cost of Revenue` /
      `average (Total) (Net) Inventory(ies), typically between two consecutive fiscal year-ends`
    )

    Adjusted Income Metric Formulas
    -------------------------------

    `(Unadjusted) Earnings before Interest, Tax, Depreciation & Amortization, a.k.a. EBITDA` = (
      `(Unadjusted) Operating Income, a.k.a. Operating Profit, or Operating Earnings (or Loss(es))` +
      `Depreciation & Amortization, a.k.a. D&A (of Fixed Assets or Property, Plant & Equipment (PP&E))`
    )

    Profitability Margin Metric Formulas
    ------------------------------------

    `Cost of Goods Sold (COGS) Margin` = (
      `(Total) Cost of Goods Sold, a.k.a. (Total) COGS, or (Total) Cost of Sales, or (Total) Cost of Revenue` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `Gross (Income or Profit or Earnings) Margin` = (
      `Gross Income, a.k.a. Gross Profit, or Gross Earnings (or Loss(es))` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `(Unadjusted) Operating (Income or Profit or Earnings) Margin` = (
      `(Unadjusted) Operating Income, a.k.a. Operating Profit, or Operating Earnings (or Loss(es))` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `(Unadjusted) EBITDA Margin` = (
      `(Unadjusted) Earnings before Interest, Tax, Depreciation & Amortization, a.k.a. EBITDA` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `Depreciation & Amortization (D&A) Margin` = (
      `Depreciation & Amortization, a.k.a. D&A (of Fixed Assets or Property, Plant & Equipment (PP&E))` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `EBIT Margin` = (
      `Earnings before Interest & Tax, a.k.a. EBIT` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `Net (Income or Profit or Earnings) Margin` = (
      `Net Income, a.k.a. Net Profit, or Net Earnings (or Loss(es)) (Attributable to Shareholders)` /
      `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    Profit-Utilization / Capital-Return Metric Formulas
    ---------------------------------------------------

    `Interest Coverage Ratio` = `Earnings before Interest & Tax, a.k.a. EBIT` / `Interest Expense`

    `Effective (Income) Tax Rate` = `(Income) Tax Expense` / `Income or Profit or Earnings (or Loss(es)) before (Income) Tax(es)`

    `Dividend Payout Ratio` = (
      `Cash Dividends` /
      `Net Income, a.k.a. Net Profit, or Net Earnings (or Loss(es)) (Attributable to Shareholders)`
    )

    `Retention Ratio` = 1 - `Dividend Payout Ratio`

    Capital-Intensiveness / Return-on-Capital Metric Formulas
    ---------------------------------------------------------

    `Capital Intensity Ratio` = `Total Assets` / `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`

    `Return on (Total) Assets, a.k.a. RoA or RoTA` = (
      `Net Income, a.k.a. Net Profit, or Net Earnings (or Loss(es)) (Attributable to Shareholders)` /
      `average Total Assets, typically between two consecutive fiscal year-ends`
    )

    Leverage Metric Formulas
    ------------------------

    `Total Debt` = (
      `Long-Term Debt (EXCLUDING any current/short-term portion)` +
      `Short-Term Debt, or Current Portion of (Long-Term) Debt`
    )

    Liquidity Metric Formulas
    -------------------------

    `(Net) Working Capital` = `(Total) Current Assets` - `(Total) Current Liabilities`

    `Working Capital Ratio` = `(Total) Current Assets` / `(Total) Current Liabilities`

    `Quick Ratio` = (
      (`Cash & Cash Equivalents` +
       `Short-Term Investments or (Current) Marketable Securities` +
       `(Net) Accounts Receivable, a.k.a. (Net) (Trade) Receivables`)
      / `(Total) Current Liabilities`
    )

    `Operating Cash Flow Ratio` = (
      `(Net) Cash Flows from Operations, a.k.a. (Net) Operating Cash Flows`
      / `(Total) Current Liabilities`
    )

    `Free Cash Flow, a.k.a. FCF` = (
      `(Net) Cash Flows from Operations, a.k.a. (Net) Operating Cash Flows` -
      `Capital Expenditure(s), a.k.a. CapEx, or Capital Spending, or Property, Plant & Equipment (PP&E) Expenditure(s)/Purchase(s)`
    )

    `Free Cash Flow Conversion Ratio` = `Free Cash Flow, a.k.a. FCF` / `Earnings before Interest, Tax, Depreciation & Amortization, a.k.a. EBITDA`

    `Days Inventory Outstanding, a.k.a. DIO` = (
      365 * `average (Total) (Net) Inventory(ies), typically between two consecutive fiscal year-ends`
      / `(Total) Cost of Goods Sold, a.k.a. (Total) COGS, or (Total) Cost of Sales, or (Total) Cost of Revenue`
    )

    `Days Payable Outstanding, a.k.a. DPO` = (
      365 * `average Accounts Payable, typically between two consecutive fiscal year-ends`
      / (`(Total) Cost of Goods Sold, a.k.a. (Total) COGS, or (Total) Cost of Sales, or (Total) Cost of Revenue` +
         `change in (Total) (Net) Inventory(ies), typically between two consecutive fiscal year-ends`)
    )

    `Days Sales Oustanding, a.k.a. DSO` = (
      365 * `average (Net) Accounts Receivable, a.k.a. (Net) (Trade) Receivables, typically between two consecutive fiscal year-ends`
      / `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
    )

    `Cash Conversion Cycle, a.k.a. CCC` = (
      `Days Inventory Outstanding, a.k.a. DIO` + `Days Sales Oustanding, a.k.a. DSO` - `Days Payable Outstanding, a.k.a. DPO`
    )

    CAPITAL-INTENSIVENESS EVALUATION
    ================================

    Capital-intensive businesses tend to have one or several of the following characteristics:

    - high `(Net) Fixed Assets, a.k.a. (Net) Property, Plant & Equipment (PP&E)` as proportion of `Total Assets`,
      e.g., over 25%;

    - high `Total Assets` relative to `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`,
      e.g., over 2 times;

    - high `Capital Expenditure(s), a.k.a. CapEx, or Capital Spending, or Property, Plant & Equipment (PP&E) Expenditure(s)/Purchase(s)`
      relative to `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`,
      e.g., over 10%;

      and/or

    - low `Return on (Total) Assets, a.k.a. RoA or RoTA`,
      e.g., under 10%

    FINANCIAL-ANALYSIS LANGUAGE/WORDING
    ===================================

    - When we see the word "average" used before a Balance-Sheet line item in financial-analysis contexts,
      it often means the simple arithmetic mean of that line item's values at two consecutive fiscal year-ends

    - When we see the word "growth" or the the phrase "top-line performance"
      used in the context of product categories/segments or geographies, unless otherwise specified,
      it often concerns the year-on-year proportional/relative growth rates in `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
      from such product categories/segments or geographies

    - When we see the word "margin" in financial-analysis contexts,
      we usually need to divide a certain Profit-and-Loss (P&L) line item by the `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales`
      for the concerned financial reporting period(s)

    - When we see the word "turnover ratio" used after a Balance-Sheet line item in financial-analysis contexts,
      we usually need to divide the `(Total) (Net) (Operating) Revenue(s), a.k.a. (Total) (Net) Sales` for the fiscal year
      by the average of that Balance-Sheet line item between the latest fiscal year-end and the immediately-preceeding fiscal year-end

    PHYSICAL-PRODUCT COMPANY PERFORMANCE METRICS vs. FINANCIAL-SERVICES COMPANY PERFORMANCE METRICS
    ===============================================================================================

    Margin metrics are generally relevant/useful performance indicators for evaluating companies making physical products,
    but generally irrelevant / not useful for evaluating companies delivering financial services

    PULICLY LISTED & TRADED DEBT SECURITIES
    =======================================
    If there are debt securities registered to trade on a national securities exchange under a company's name,
    then such debt securities are enumerated on SEC filings' 1st page, alongside the company's public equity shares/stock.

    If only equity shares/stock are enumerated on SEC filings' 1st page,
    then there are no debt securities registered to trade on a national securities exchange under a company's name.
"""

_FEW_SHOT_INSTRUCTIONS = """
    Consider the examples below.
    {few_shot_examples}
"""


PROMPT_TEMPLATES = {
    "default": {
        "instructions": _DEFAULT_INSTRUCTIONS,
        "context": _CONTEXT,
        "query_str": _QUERY_STR,
        "few_shot_examples": _FEW_SHOT_INSTRUCTIONS,
    },
    "concise": {
        "instructions": _CONCISE_INSTRUCTIONS,
        "context": _CONTEXT,
        "query_str": _QUERY_STR,
        "few_shot_examples": _FEW_SHOT_INSTRUCTIONS,
    },
    "CoT": {
        "instructions": _COT_INSTRUCTIONS,
        "context": _CONTEXT,
        "query_str": _QUERY_STR,
        "few_shot_examples": _FEW_SHOT_INSTRUCTIONS,
    },
    "finance-expert": {
        "instructions": _FINANCE_EXPERT_INSTRUCTIONS,
        "context": _CONTEXT,
        "query_str": _QUERY_STR,
        "few_shot_examples": _FEW_SHOT_INSTRUCTIONS,
    },
}

MAIN_LAYOUT = RichPromptTemplate("""
    {{instructions}}
    {% if with_context %} {{context}} {% endif %}
    {% if with_few_shot_prompt %} {{few_shot_examples}} {% endif %}
    {{query_str}}
""")


def get_template(
    template_name: str, with_context: bool = False, with_few_shot_prompt: bool = False
) -> str:
    """Returns a formatted prompt specified by a template name."""
    components = PROMPT_TEMPLATES[template_name]
    components["with_context"] = with_context  # type: ignore
    components["with_few_shot_prompt"] = with_few_shot_prompt  # type: ignore
    return MAIN_LAYOUT.format(**components)  # type: ignore


def get_template_names() -> T.List[str]:
    """Returns all template names."""
    return list(PROMPT_TEMPLATES.keys())


def _load_json_file(file_path):
    logger.debug(f"Loading: {file_path}")
    with open(file_path, "r") as file:
        data = json.load(file)
    return data


def get_agent_template(prompt_name: str):
    agentic_templates = _load_json_file(cfg.paths.agentic_templates)
    return agentic_templates[prompt_name]



import inspect
import json
import time
from typing import Any, Dict, Optional, Type, Union

import google
from azure.ai.inference.models._models import ChatCompletionsToolCall
from llama_index.core.base.llms.types import ChatResponse, CompletionResponse
from llama_index.core.evaluation import CorrectnessEvaluator
from llama_index.core.instrumentation.events import BaseEvent
from llama_index.core.instrumentation.events.llm import (
    LLMChatEndEvent,
    LLMChatStartEvent,
    LLMCompletionEndEvent,
    LLMCompletionStartEvent,
)
from openinference.instrumentation.llama_index._handler import (
    _SUPPRESS_INSTRUMENTATION_KEY,
    INPUT_VALUE,
    JSON,
    LLM_INVOCATION_PARAMETERS,
    LLM_MODEL_NAME,
    LLM_TOKEN_COUNT_COMPLETION,
    LLM_TOKEN_COUNT_PROMPT,
    LLM_TOKEN_COUNT_TOTAL,
    OUTPUT_MIME_TYPE,
    OUTPUT_VALUE,
    EventHandler,
    _get_token_counts,
    _init_span_kind,
    _Span,
    _SpanHandler,
    context_api,
    get_attributes_from_context,
    time_ns,
)
from openinference.semconv.trace import OpenInferenceMimeTypeValues
from opentelemetry.trace import NoOpTracer
from pydantic import BaseModel, PrivateAttr
from vertexai.preview.tokenization import get_tokenizer_for_model

from syftr.configuration import LLMCostCharacters, LLMCostHourly, LLMCostTokens, cfg
from syftr.logger import logger

# Unit: $ / token
MODEL_PRICING_INFO: Dict[str, Dict[str, Dict[str, float]]] = {
    "tokens": {
        "gpt-4o-mini": {"input": 0.15 / 1e6, "output": 0.6 / 1e6},
        "gpt-4o": {"input": 2.5 / 1e6, "output": 10.0 / 1e6},
        "o1": {"input": 15.0 / 1e6, "output": 60.0 / 1e6},
        "o3-mini": {"input": 3.0 / 1e6, "output": 12 / 1e6},
        "gpt-35-turbo": {"input": 1.00 / 1e6, "output": 2.00 / 1e6},
        "mistral-large-2411": {"input": 3.00 / 1e6, "output": 9.00 / 1e6},
        "Llama-3.3-70B-Instruct": {"input": 0.71 / 1e6, "output": 0.71 / 1e6},
        "Phi-4": {"input": 0.125 / 1e6, "output": 0.50 / 1e6},
        "claude-3-5-sonnet-v2@20241022": {"input": 3.00 / 1e6, "output": 15.0 / 1e6},
        "claude-3-5-haiku@20241022": {"input": 0.80 / 1e6, "output": 4.00 / 1e6},
        # Azure costs not released yet - using Together.AI pricing for now
        "Deepseek-R1": {"input": 7.00 / 1e6, "output": 7.00 / 1e6},
        "deepseek-ai/DeepSeek-R1": {"input": 7.00 / 1e6, "output": 7.00 / 1e6},
        "deepseek-ai/DeepSeek-V3": {"input": 1.25 / 1e6, "output": 1.25 / 1e6},
        "gemini-2.0-flash-lite-preview-02-05": {
            "input": 0.075 / 1e6,
            "output": 0.3 / 1e6,
        },
        "gemini-flash-think-exp": {"input": 0.15 / 1e6, "output": 0.60 / 1e6},
        "gemini-flash-lite-preview-02-05": {"input": 0.075 / 1e6, "output": 0.3 / 1e6},
        "gemini-2.0-pro-exp-02-05": {"input": 1.50 / 1e6, "output": 1.50 / 1e6},
        "gemini-2.0-flash-001": {"input": 0.15 / 1e6, "output": 0.60 / 1e6},
        # New Gemini prices not yet available
        # "gemini-2.0-flash-thinking-exp-01-21": {
        #     "input": 0.15 / 1e6,
        #     "output": 0.60 / 1e6,
        # },
        # Pricing not yet known/available
        # "llama3.1-8b": {"input": 1.50 / 1e6, "output": 1.50 / 1e6},
        # "llama-3.3-70b": {"input": 2.68 / 1e6, "output": 3.54 / 1e6},
        "google/gemma-3-27b-it": {"input": 0.80 / 1e6, "output": 0.80 / 1e6},
        "nvidia/Llama-3_3-Nemotron-Super-49B": {
            "input": 0.90 / 1e6,
            "output": 0.90 / 1e6,
        },
        # Local models use together.ai pricing structure
        "Qwen/Qwen2.5": {"input": 0.80 / 1e6, "output": 0.80 / 1e6},
        "Qwen/Qwen3-32B": {"input": 0.80 / 1e6, "output": 0.80 / 1e6},
        "deepseek-ai/DeepSeek-R1-Distill-Llama-70B": {
            "input": 0.88 / 1e6,
            "output": 0.88 / 1e6,
        },
        "microsoft/Phi-4-multimodal-instruct": {
            "input": 0.2 / 1e6,
            "output": 0.2 / 1e6,
        },
        "llama3.1-8b": {
            "input": 0.2 / 1e6,
            "output": 0.2 / 1e6,
        },
        "cerebras-llama33-70B": {
            "input": 0.85 / 1e6,
            "output": 1.2 / 1e6,
        },
        "cerebras-qwen-3": {
            "input": 0.4 / 1e6,
            "output": 0.8 / 1e6,
        },
        "cerebras-scout": {
            "input": 0.65 / 1e6,
            "output": 0.85 / 1e6,
        },
        "cerebras-llama31-8B": {
            "input": 0.1 / 1e6,
            "output": 0.1 / 1e6,
        },
        "cerebras-deepseek": {
            "input": 2.2 / 1e6,
            "output": 2.5 / 1e6,
        },
    },
    "characters": {
        "gemini-1.5-pro-002": {"input": 0.0003125 / 1e3, "output": 0.00125 / 1e3},
        "gemini-1.5-flash-002": {"input": 0.00001875 / 1e3, "output": 0.000075 / 1e3},
    },
    "seconds": {
        "datarobot/DeepSeek-Llama": {"cost_per_second": 12.06654 / 3600},
    },
}

if cfg.generative_models:
    logger.debug(
        f"Updating MODEL_PRICING_INFO from cfg.generative_models: {list(cfg.generative_models.keys())}"
    )
    for config_key, llm_config_item in cfg.generative_models.items():
        model_key = llm_config_item.model_name
        cost_config = llm_config_item.cost
        cost_type = getattr(cost_config, "type", "unknown")

        if cost_type == "tokens":
            assert isinstance(cost_config, LLMCostTokens), cost_config
            # Ensure the sub-dictionary for 'tokens' exists
            MODEL_PRICING_INFO["tokens"].setdefault(model_key, {})
            # cost_config.input/output are "Cost per million tokens"
            # MODEL_PRICING_INFO stores "Cost per token"
            MODEL_PRICING_INFO["tokens"][model_key]["input"] = (
                cost_config.input / 1_000_000.0
            )
            MODEL_PRICING_INFO["tokens"][model_key]["output"] = (
                cost_config.output / 1_000_000.0
            )
            logger.debug(
                f"Updated token pricing for '{model_key}' from cfg ('{config_key}')."
            )
        elif cost_type == "characters":
            assert isinstance(cost_config, LLMCostCharacters), cost_config
            MODEL_PRICING_INFO["characters"].setdefault(model_key, {})
            # cost_config.input/output are "Cost per million characters"
            # MODEL_PRICING_INFO stores "Cost per character"
            MODEL_PRICING_INFO["characters"][model_key]["input"] = (
                cost_config.input / 1_000_000.0
            )
            MODEL_PRICING_INFO["characters"][model_key]["output"] = (
                cost_config.output / 1_000_000.0
            )
            logger.debug(
                f"Updated character pricing for '{model_key}' from cfg ('{config_key}')."
            )
        elif cost_type == "hourly":
            assert isinstance(cost_config, LLMCostHourly), cost_config
            MODEL_PRICING_INFO["seconds"].setdefault(model_key, {})
            # cost_config.rate is "Average inference cost per hour"
            # MODEL_PRICING_INFO stores "Cost per second"
            MODEL_PRICING_INFO["seconds"][model_key]["cost_per_second"] = (
                cost_config.rate / 3600.0
            )
            logger.debug(
                f"Updated hourly (per second) pricing for '{model_key}' from cfg ('{config_key}')."
            )
        else:
            raise ValueError(
                f"Unknown or unsupported cost type '{cost_type}' for model '{model_key}' "
                f"(from cfg key '{config_key}'). Pricing not updated."
            )


class LLMCallData(BaseModel):
    llm_name: str
    input_tokens: int
    output_tokens: int
    input_characters: int
    output_characters: int
    total_tokens: int
    llm_call_latency: float
    invocation_parameters: str | None = None

    @property
    def cost(self) -> float:
        if price_info := MODEL_PRICING_INFO["tokens"].get(self.llm_name):
            return (
                self.input_tokens * price_info["input"]
                + self.output_tokens * price_info["output"]
            )

        if price_info := MODEL_PRICING_INFO["characters"].get(self.llm_name):
            return (
                self.input_characters * price_info["input"]
                + self.output_characters * price_info["output"]
            )

        if price_info := MODEL_PRICING_INFO["seconds"].get(self.llm_name):
            return price_info["cost_per_second"] * self.llm_call_latency

        raise ValueError(f"Don't know how to compute costs for LLM {self.llm_name=}")


class CircularReferenceEncoder(json.JSONEncoder):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.seen = set()

    def default(self, obj):
        if id(obj) in self.seen:
            return f"<Circular reference to {type(obj).__name__}>"
        self.seen.add(id(obj))
        if isinstance(obj, google.cloud.aiplatform_v1beta1.types.tool.FunctionCall):
            return obj.to_dict()
        if isinstance(obj, ChatCompletionsToolCall):
            return obj.as_dict()
        return super().default(obj)


class TokenTrackingSpan(_Span):
    _instance: Any = PrivateAttr()
    _bound_args: inspect.BoundArguments = PrivateAttr()

    def __init__(
        self,
        *args,
        instance: Any,
        bound_args: inspect.BoundArguments,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._instance = instance
        self._bound_args = bound_args

    def process_event(self, event: BaseEvent, **kwargs: Any) -> Any:
        # For LLM events, the super class will attach token counts and
        # to the span (this object)
        # Invocation params and model name should have already been attached
        # by self.process_instance on span creation
        super().process_event(event)

        if isinstance(event, (LLMChatStartEvent, LLMCompletionStartEvent)):
            self["call_start"] = time.time()

        if not isinstance(event, (LLMChatEndEvent, LLMCompletionEndEvent)):
            return

        self["call_end"] = time.time()
        latency = self._attributes["call_end"] - self._attributes["call_start"]  # type: ignore

        logger.debug("Model input: %s", self._attributes[INPUT_VALUE])
        logger.debug("Model output: %s", self._attributes[OUTPUT_VALUE])

        call_data = LLMCallData(
            llm_name=self._attributes[LLM_MODEL_NAME],  # type: ignore
            input_tokens=self._attributes[LLM_TOKEN_COUNT_PROMPT],  # type: ignore
            output_tokens=self._attributes[LLM_TOKEN_COUNT_COMPLETION],  # type: ignore
            total_tokens=self._attributes[LLM_TOKEN_COUNT_TOTAL],  # type: ignore
            input_characters=len(self._attributes[INPUT_VALUE]),  # type: ignore
            output_characters=len(self._attributes[OUTPUT_VALUE]),  # type: ignore
            invocation_parameters=self._attributes[LLM_INVOCATION_PARAMETERS],  # type: ignore
            llm_call_latency=latency,
        )
        # Recursively get the parent span until we find the Flow instance
        root_flow_span = self._find_root_flow_span()
        if (
            invocation_id := root_flow_span._bound_args.arguments.get("invocation_id")
        ) is not None:
            # Store the event data in the Flow instance
            assert hasattr(root_flow_span._instance, "_llm_call_data"), (
                root_flow_span._instance
            )
            root_flow_span._instance._llm_call_data[invocation_id].append(call_data)
        return event

    def _has_token_counts(self) -> bool:
        return all(
            key in self._attributes
            for key in [
                LLM_TOKEN_COUNT_PROMPT,
                LLM_TOKEN_COUNT_COMPLETION,
                LLM_TOKEN_COUNT_TOTAL,
            ]
        )

    def _extract_token_counts(
        self, response: Union[ChatResponse, CompletionResponse]
    ) -> None:
        """Add additional response handling logic for additional LLM response schemas.

        Called by super().process_event() for LLM Chat/Completion End events
        """
        super()._extract_token_counts(response)
        if self._has_token_counts():
            # Successfully extracted counts
            return

        # OpenAI Responses API
        if (
            (raw := getattr(response, "raw", None))
            and (usage := getattr(raw, "usage", None))
            and (input_tokens := getattr(usage, "input_tokens", None)) is not None
            and (output_tokens := getattr(usage, "output_tokens", None)) is not None
            and (total_tokens := getattr(usage, "total_tokens", None)) is not None
        ):
            self[LLM_TOKEN_COUNT_PROMPT] = input_tokens
            self[LLM_TOKEN_COUNT_COMPLETION] = output_tokens
            self[LLM_TOKEN_COUNT_TOTAL] = total_tokens
            return

        # GPT models on AzureOpenAI
        if (raw := getattr(response, "raw", None)) and (
            usage := getattr(raw, "usage", None)
        ):
            for k, v in _get_token_counts(usage):
                self[k] = v
            return

        # Claude Models on GCP
        if (
            (raw := getattr(response, "raw", None))
            and hasattr(raw, "get")
            and (usage := raw.get("usage"))
            and (input_tokens := getattr(usage, "input_tokens", None))
            and (output_tokens := getattr(usage, "output_tokens", None))
            and not getattr(usage, "total_tokens", None)
        ):
            self[LLM_TOKEN_COUNT_PROMPT] = input_tokens
            self[LLM_TOKEN_COUNT_COMPLETION] = output_tokens
            self[LLM_TOKEN_COUNT_TOTAL] = input_tokens + output_tokens
            return

        # GCP Gemini responses (old)
        if (
            (raw := getattr(response, "raw", None))
            and hasattr(raw, "get")
            and (raw_resp := raw.get("_raw_response"))
            and (usage := getattr(raw_resp, "usage_metadata", None))
            and (input_tokens := getattr(usage, "prompt_token_count", None))
            and (output_tokens := getattr(usage, "candidates_token_count", None))
            and (total_tokens := getattr(usage, "total_token_count", None))
        ):
            self[LLM_TOKEN_COUNT_PROMPT] = input_tokens
            self[LLM_TOKEN_COUNT_COMPLETION] = output_tokens
            self[LLM_TOKEN_COUNT_TOTAL] = total_tokens
            return

        if self._attributes[LLM_MODEL_NAME] in MODEL_PRICING_INFO["characters"].keys():
            num_input_tokens = len(str(self._attributes[INPUT_VALUE]))
            num_output_tokens = len(str(self._attributes[OUTPUT_VALUE]))
            self[LLM_TOKEN_COUNT_PROMPT] = num_input_tokens
            self[LLM_TOKEN_COUNT_COMPLETION] = num_output_tokens
            self[LLM_TOKEN_COUNT_TOTAL] = num_input_tokens + num_output_tokens
            return

        # GCP Gemini responses (new)
        if self._attributes[LLM_MODEL_NAME] in MODEL_PRICING_INFO[
            "tokens"
        ].keys() and "gemini" in str(self._attributes[LLM_MODEL_NAME]):
            try:
                try:
                    tokenizer = get_tokenizer_for_model(
                        self._attributes[LLM_MODEL_NAME]
                    )
                except Exception as e:
                    logger.debug("Vertex tokenizer error: %s", e)
                    # temporary hack
                    tokenizer = get_tokenizer_for_model("gemini-1.5-flash-002")
                num_input_tokens = tokenizer.count_tokens(
                    self._attributes[INPUT_VALUE]
                ).total_tokens
                num_output_tokens = tokenizer.count_tokens(
                    self._attributes[OUTPUT_VALUE]
                ).total_tokens
                self[LLM_TOKEN_COUNT_PROMPT] = num_input_tokens
                self[LLM_TOKEN_COUNT_COMPLETION] = num_output_tokens
                self[LLM_TOKEN_COUNT_TOTAL] = num_input_tokens + num_output_tokens
                return
            except Exception as e:
                logger.warning("Vertex tokenizer error: %s", e)

        raise ValueError(
            f"Failed to extract token counts from response object `{response=}` when using model `{self._attributes[LLM_MODEL_NAME]}`"
        )

    def _find_root_flow_span(self) -> "TokenTrackingSpan":
        parent_span = self._parent
        if parent_span is None:
            return self
        last_parent = parent_span
        while parent_span is not None:
            assert isinstance(parent_span, TokenTrackingSpan), parent_span
            instance = getattr(parent_span, "_instance", None)
            if _is_flow(instance) or _is_evaluator(instance):
                return parent_span
            last_parent = parent_span
            parent_span = parent_span._parent
        raise ValueError(
            f"Could not find Flow class in parent spans of `{self=}`. Root span is `{last_parent}`"
        )

    def model_dump_json(self, **kwargs):
        return json.dumps(self.dict(**kwargs), cls=CircularReferenceEncoder)

    def process_output(self, instance, result):
        try:
            self[OUTPUT_VALUE] = json.dumps(
                result.dict(exclude_unset=True), cls=CircularReferenceEncoder
            )
            self[OUTPUT_MIME_TYPE] = JSON
        except Exception as e:
            logger.error(f"Error serializing result to JSON: {e}")
            # Fallback: convert to string
            self[OUTPUT_VALUE] = str(result)
            self[OUTPUT_MIME_TYPE] = OpenInferenceMimeTypeValues.TEXT.value


def _is_flow(instance: Any) -> bool:
    return any("Flow" in cls.__name__ for cls in instance.__class__.mro())


def _is_evaluator(instance: Any) -> bool:
    return isinstance(instance, CorrectnessEvaluator)


def init_span_kind(instance: Any) -> Optional[str]:
    if _is_flow(instance):
        return "SyftrFlow"
    if _is_evaluator(instance):
        return "SyftrCorrectnessEvaluator"
    return _init_span_kind(instance)


def patch_pydantic_instance(
    instance: Any,
    un_serializable_type: Type,
    *fields: str,
    to_dict_method_name: str = "to_dict",
):
    """
    Patches a Pydantic BaseModel instance to add a custom serializer
    for fields of a specific type.

    Args:
        instance: The Pydantic BaseModel instance to patch.
        un_serializable_type: The class of the type that needs custom serialization.
        to_dict_method_name: The name of the method to call on the
            unserializable object to convert it to a dictionary (default: "to_dict")
    """

    if not isinstance(instance, BaseModel):
        return instance

    for attr, value in instance.model_dump().items():
        if isinstance(value, un_serializable_type):
            setattr(instance, attr, getattr(value, to_dict_method_name)())

    return instance


class TokenTrackingSpanHandler(_SpanHandler):
    def new_span(
        self,
        id_: str,
        bound_args: inspect.BoundArguments,
        instance: Optional[Any] = None,
        parent_span_id: Optional[str] = None,
        tags: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Optional[_Span]:
        """Copied from the base class, with custom span type and args."""
        if context_api.get_value(_SUPPRESS_INSTRUMENTATION_KEY):
            return None
        with self.lock:
            parent = self.open_spans.get(parent_span_id) if parent_span_id else None
        otel_span = self._otel_tracer.start_span(
            name=id_.partition("-")[0],
            start_time=time_ns(),
            attributes=dict(get_attributes_from_context()),
            context=(parent.context if parent else None),
        )
        span = TokenTrackingSpan(
            otel_span=otel_span,
            span_kind=_init_span_kind(instance),
            parent=parent,
            id_=id_,
            parent_id=parent_span_id,
            instance=instance,
            bound_args=bound_args,
        )
        span.process_instance(instance)
        span.process_input(instance, bound_args)
        return span

    def prepare_to_exit_span(
        self,
        id_: str,
        bound_args: inspect.BoundArguments,
        instance: Optional[Any] = None,
        result: Optional[Any] = None,
        **kwargs: Any,
    ) -> Any:
        result = patch_pydantic_instance(
            result, google.cloud.aiplatform_v1beta1.types.tool.FunctionCall
        )
        return super().prepare_to_exit_span(id_, bound_args, instance, result, **kwargs)


class TokenTrackingEventHandler(EventHandler):
    """Subclass of OpenInference handler."""

    def __init__(self, **kwargs) -> None:
        super().__init__(tracer=NoOpTracer())
        # Overwrite _span_handler with modified span handler
        self._span_handler = TokenTrackingSpanHandler(tracer=NoOpTracer())  # type: ignore



import logging
import typing as T

import pytest

from syftr.embeddings.timeouts import EmbeddingPreemptiveTimeoutError
from syftr.huggingface_helper import get_embedding_model
from syftr.logger import logger
from syftr.storage import FinanceBenchHF, PartitionMap
from syftr.studies import StudyConfig


def test_embedding_models() -> None:
    study_config = StudyConfig(
        name="test-embedding-models",
        dataset=FinanceBenchHF(
            partition_map=PartitionMap(test="pepsi"),
        ),
    )
    search_space = study_config.search_space

    for name in search_space.rag_retriever.embedding_models:
        logger.info(f"Testing embedding model: {name}")
        model, is_onnx = get_embedding_model(
            name,
            study_config.timeouts,
            total_chunks=0,
            use_hf_endpoint_models=False,
        )
        assert is_onnx in [True, False], "Bad is_onnx response"
        assert model is not None, f"Model for '{name}' is None"
        out: T.List[float] = model.get_query_embedding("Do you function?")
        assert out, f"Cannot generate embeddings with '{name}'"


def test_embedding_model_timeout(build_index_that_will_time_out) -> None:
    with pytest.raises(EmbeddingPreemptiveTimeoutError):
        build_index_that_will_time_out()


def test_embedding_model_no_timeout(build_index_that_will_not_time_out) -> None:
    build_index_that_will_not_time_out()


def test_embedding_model_skip_timeout(
    caplog, build_index_that_will_skip_time_out
) -> None:
    with caplog.at_level(logging.WARNING):
        index, _ = build_index_that_will_skip_time_out()
        assert "Not raising embedding timeout because it is deactivated" in caplog.text
        assert index._embed_model._warning_logged



import json
import os
import typing as T
from json import JSONDecodeError

import tiktoken
from anthropic import AnthropicVertex, AsyncAnthropicVertex
from google.cloud.aiplatform_v1beta1.types import content
from google.oauth2 import service_account

# from llama_index.core.base.llms.types import LLMMetadata, MessageRole
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.llms.llm import LLM

# from llama_index.embeddings.azure_openai import AzureOpenAIEmbedding
from llama_index.llms.anthropic import Anthropic
from llama_index.llms.azure_inference import AzureAICompletionsModel
from llama_index.llms.azure_openai import AzureOpenAI
from llama_index.llms.cerebras import Cerebras
from llama_index.llms.openai import OpenAIResponses
from llama_index.llms.openai_like import OpenAILike
from llama_index.llms.vertex import Vertex
from mypy_extensions import DefaultNamedArg

from syftr.configuration import (
    NON_OPENAI_CONTEXT_WINDOW_FACTOR,
    AnthropicVertexLLM,
    AzureAICompletionsLLM,
    AzureOpenAILLM,
    CerebrasLLM,
    OpenAILikeLLM,
    OpenAIResponsesLLM,
    Settings,
    VertexAILLM,
    cfg,
)
from syftr.logger import logger
from syftr.patches import _get_all_kwargs, _get_model_kwargs_responses

BASELINE_RAG_EMBEDDING_MODEL = "BAAI/bge-small-en-v1.5"

Anthropic._get_all_kwargs = _get_all_kwargs  # type: ignore


def _scale(
    context_window_length: int, factor: float = NON_OPENAI_CONTEXT_WINDOW_FACTOR
) -> int:
    return int(context_window_length * factor)


if (hf_token := cfg.hf_embeddings.api_key.get_secret_value()) != "NOT SET":
    os.environ["HF_TOKEN"] = hf_token

GCP_SAFETY_SETTINGS = {
    content.HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: content.SafetySetting.HarmBlockThreshold.BLOCK_ONLY_HIGH,
    content.HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: content.SafetySetting.HarmBlockThreshold.BLOCK_ONLY_HIGH,
    content.HarmCategory.HARM_CATEGORY_HARASSMENT: content.SafetySetting.HarmBlockThreshold.BLOCK_ONLY_HIGH,
    content.HarmCategory.HARM_CATEGORY_HATE_SPEECH: content.SafetySetting.HarmBlockThreshold.BLOCK_ONLY_HIGH,
}

try:
    GCP_CREDS = json.loads(cfg.gcp_vertex.credentials.get_secret_value())
except JSONDecodeError:
    GCP_CREDS = {}


def add_scoped_credentials_anthropic(anthropic_llm: Anthropic) -> Anthropic:
    """Add Google service account credentials to an Anthropic LLM"""
    credentials = (
        service_account.Credentials.from_service_account_info(GCP_CREDS).with_scopes(
            ["https://www.googleapis.com/auth/cloud-platform"]
        )
        if GCP_CREDS
        else None
    )
    sync_client = anthropic_llm._client
    assert isinstance(sync_client, AnthropicVertex)
    sync_client.credentials = credentials
    anthropic_llm._client = sync_client
    async_client = anthropic_llm._aclient
    assert isinstance(async_client, AsyncAnthropicVertex)
    async_client.credentials = credentials
    anthropic_llm._aclient = async_client
    return anthropic_llm


def _construct_azure_openai_llm(name: str, llm_config: AzureOpenAILLM) -> AzureOpenAI:
    llm_config.additional_kwargs = llm_config.additional_kwargs or {}
    return AzureOpenAI(
        model=llm_config.model_name,
        temperature=llm_config.temperature,
        top_p=llm_config.top_p,
        max_tokens=llm_config.max_tokens,
        max_retries=0,
        system_prompt=llm_config.system_prompt,
        engine=llm_config.deployment_name,
        api_key=llm_config.api_key.get_secret_value()
        if llm_config.api_key
        else cfg.azure_oai.api_key.get_secret_value(),
        azure_endpoint=llm_config.api_url.unicode_string()
        if llm_config.api_url
        else cfg.azure_oai.api_url.unicode_string(),
        api_version=llm_config.api_version or cfg.azure_oai.api_version,
        additional_kwargs=llm_config.additional_kwargs or {},
    )


def _construct_vertex_ai_llm(name: str, llm_config: VertexAILLM) -> Vertex:
    credentials = (
        service_account.Credentials.from_service_account_info(GCP_CREDS)
        if GCP_CREDS
        else {}
    )
    return Vertex(
        model=llm_config.model_name,
        temperature=llm_config.temperature,
        max_tokens=llm_config.max_tokens,
        max_retries=0,
        system_prompt=llm_config.system_prompt,
        project=llm_config.project_id or cfg.gcp_vertex.project_id,
        location=llm_config.region or cfg.gcp_vertex.region,
        safety_settings=llm_config.safety_settings or GCP_SAFETY_SETTINGS,
        credentials=credentials,
        context_window=llm_config.context_window,
        additional_kwargs=llm_config.additional_kwargs or {},
    )


def _construct_anthropic_vertex_llm(
    name: str, llm_config: AnthropicVertexLLM
) -> Anthropic:
    anthropic_llm = Anthropic(
        model=llm_config.model_name,
        temperature=llm_config.temperature,
        max_tokens=llm_config.max_tokens,
        max_retries=0,
        system_prompt=llm_config.system_prompt,
        project_id=llm_config.project_id or cfg.gcp_vertex.project_id,
        region=llm_config.region or cfg.gcp_vertex.region,
        thinking_dict=llm_config.thinking_dict,
        additional_kwargs=llm_config.additional_kwargs or {},
    )
    return add_scoped_credentials_anthropic(anthropic_llm)


def _construct_azure_ai_completions_llm(
    name: str, llm_config: AzureAICompletionsLLM
) -> AzureAICompletionsModel:
    return AzureAICompletionsModel(
        model_name=llm_config.model_name,
        temperature=llm_config.temperature,
        max_tokens=llm_config.max_tokens,
        system_prompt=llm_config.system_prompt,
        endpoint=llm_config.api_url.unicode_string(),
        credential=llm_config.api_key.get_secret_value(),
        client_kwargs=llm_config.client_kwargs,
        api_version=llm_config.api_version,
    )


def _construct_cerebras_llm(name: str, llm_config: CerebrasLLM) -> Cerebras:
    return Cerebras(
        model=llm_config.model_name,
        temperature=llm_config.temperature,
        max_tokens=llm_config.max_tokens,
        max_retries=0,
        system_prompt=llm_config.system_prompt,
        api_key=cfg.cerebras.api_key.get_secret_value(),
        api_base=cfg.cerebras.api_url.unicode_string(),
        context_window=llm_config.context_window,  # Use raw value as per existing Cerebras configs
        is_chat_model=llm_config.is_chat_model,
        is_function_calling_model=llm_config.is_function_calling_model,
        additional_kwargs=llm_config.additional_kwargs or {},
    )


def _construct_openai_like_llm(name: str, llm_config: OpenAILikeLLM) -> OpenAILike:
    return OpenAILike(
        model=llm_config.model_name,
        temperature=llm_config.temperature,
        max_tokens=llm_config.max_tokens,
        max_retries=0,
        system_prompt=llm_config.system_prompt,
        api_base=str(llm_config.api_base),
        api_key=llm_config.api_key.get_secret_value(),
        api_version=llm_config.api_version,  # type: ignore
        context_window=llm_config.context_window,
        is_chat_model=llm_config.is_chat_model,
        is_function_calling_model=llm_config.is_function_calling_model,
        timeout=llm_config.timeout,
        additional_kwargs=llm_config.additional_kwargs or {},
    )


def _construct_openai_responses_llm(
    name: str, llm_config: OpenAIResponsesLLM
) -> OpenAIResponses:
    OpenAIResponses._get_model_kwargs = _get_model_kwargs_responses
    return OpenAIResponses(
        model=llm_config.model_name,
        temperature=llm_config.temperature,
        max_output_tokens=llm_config.max_tokens,
        max_retries=0,
        system_prompt=llm_config.system_prompt,
        api_base=str(llm_config.api_base),
        api_key=llm_config.api_key.get_secret_value(),
        api_version=llm_config.api_version,  # type: ignore
        context_window=llm_config.context_window,
        timeout=llm_config.timeout,
        additional_kwargs=llm_config.additional_kwargs or {},
    )


LLM_NAMES__LOCAL_MODELS: T.List[str] = [
    model.model_name for model in cfg.local_models.generative or []
]
LLM_NAMES__GENERATIVE_MODELS: T.List[str] = [
    model for model in cfg.generative_models.keys()
]


if LLM_NAMES__GENERATIVE_MODELS and "gpt-4o-mini" not in LLM_NAMES__GENERATIVE_MODELS:
    BASELINE_LLM = LLM_NAMES__GENERATIVE_MODELS[0]

LLM_NAMES: T.List[str] = LLM_NAMES__LOCAL_MODELS + LLM_NAMES__GENERATIVE_MODELS
assert len(LLM_NAMES) == len(set(LLM_NAMES)), (
    "Duplicate LLM names found in configuration. Please ensure all LLM names are unique."
)

# at least one model is required for unit testing
BASELINE_LLM = "gpt-4o-mini"
if BASELINE_LLM not in LLM_NAMES:
    if LLM_NAMES:
        BASELINE_LLM = LLM_NAMES[0]
    else:
        LLM_NAMES = LLM_NAMES__GENERATIVE_MODELS = [BASELINE_LLM]


def get_generative_llm(
    name: str,
    temperature: float | None,
    top_p: float | None,
) -> FunctionCallingLLM:
    if name not in LLM_NAMES__GENERATIVE_MODELS:
        raise ValueError(
            f"LLM '{name}' not found in configured generative models. "
            f"Available generative models: {LLM_NAMES__GENERATIVE_MODELS}"
        )
    llm_instance: T.Optional[FunctionCallingLLM] = None
    llm_config_instance = cfg.generative_models[name]

    if temperature is not None:
        llm_config_instance.temperature = temperature
    if top_p is not None:
        llm_config_instance.top_p = top_p

    provider = getattr(llm_config_instance, "provider", None)

    if provider == "azure_openai" and isinstance(llm_config_instance, AzureOpenAILLM):
        llm_instance = _construct_azure_openai_llm(name, llm_config_instance)
    elif provider == "vertex_ai" and isinstance(llm_config_instance, VertexAILLM):
        llm_instance = _construct_vertex_ai_llm(name, llm_config_instance)
    elif provider == "anthropic_vertex" and isinstance(
        llm_config_instance, AnthropicVertexLLM
    ):
        llm_instance = _construct_anthropic_vertex_llm(name, llm_config_instance)
    elif provider == "azure_ai" and isinstance(
        llm_config_instance, AzureAICompletionsLLM
    ):
        llm_instance = _construct_azure_ai_completions_llm(name, llm_config_instance)
    elif provider == "cerebras" and isinstance(llm_config_instance, CerebrasLLM):
        llm_instance = _construct_cerebras_llm(name, llm_config_instance)
    elif provider == "openai_like" and isinstance(llm_config_instance, OpenAILikeLLM):
        llm_instance = _construct_openai_like_llm(name, llm_config_instance)
    elif provider == "openai_responses" and isinstance(
        llm_config_instance, OpenAIResponsesLLM
    ):
        llm_instance = _construct_openai_responses_llm(name, llm_config_instance)
    else:
        raise ValueError(
            f"Unsupported provider type '{provider}' or "
            f"mismatched Pydantic config model type for model '{name}'."
        )
    return llm_instance


def get_local_llm(
    name: str,
    temperature: float | None = None,
    top_p: float | None = None,
) -> FunctionCallingLLM:
    assert name in LLM_NAMES__LOCAL_MODELS, (
        f"LLM '{name}' not found in configured local models. "
        f"Available local models: {LLM_NAMES__LOCAL_MODELS}"
    )
    model = next(
        model for model in cfg.local_models.generative or [] if model.model_name == name
    )
    if top_p is not None:
        model.additional_kwargs["top_p"] = top_p
    return OpenAILike(  # type: ignore
        api_base=str(model.api_base),
        api_key=model.api_key.get_secret_value()
        if model.api_key is not None
        else cfg.local_models.default_api_key.get_secret_value(),
        model=model.model_name,
        temperature=temperature if temperature is not None else model.temperature,
        max_tokens=model.max_tokens,
        context_window=_scale(model.context_window),
        is_chat_model=model.is_chat_model,
        is_function_calling_model=model.is_function_calling_model,
        timeout=model.timeout,
        max_retries=model.max_retries,
        additional_kwargs=model.additional_kwargs,
    )


def get_llm(
    name: str,
    temperature: float | None = None,
    top_p: float | None = None,
) -> FunctionCallingLLM:
    if name in LLM_NAMES__GENERATIVE_MODELS:
        return get_generative_llm(name, temperature, top_p)
    if name in LLM_NAMES__LOCAL_MODELS:
        return get_local_llm(name, temperature, top_p)
    raise ValueError(
        f"LLM '{name}' not found in configured LLMs. Available LLMs: {LLM_NAMES}"
    )


def load_configured_llms(config: Settings) -> T.Dict[str, FunctionCallingLLM]:
    _dynamically_loaded_llms: T.Dict[str, FunctionCallingLLM] = {}
    if not config.generative_models:
        return {}
    logger.debug(
        f"Loading LLMs from 'generative_models' configuration: {list(config.generative_models.keys())}"
    )
    for name, llm_config_instance in config.generative_models.items():
        llm_instance: T.Optional[FunctionCallingLLM] = None
        try:
            llm_instance = get_llm(
                name=name,
                temperature=llm_config_instance.temperature,
                top_p=llm_config_instance.top_p,
            )
            if llm_instance:
                _dynamically_loaded_llms[name] = llm_instance
                logger.debug(f"Successfully loaded LLM '{name}' from configuration.")
        except Exception as e:
            # Log with traceback for easier debugging
            logger.error(
                f"Failed to load configured LLM '{name}' due to: {e}", exc_info=True
            )
            raise
    return _dynamically_loaded_llms


def get_all_llms_with_defaults() -> T.List[FunctionCallingLLM]:
    return [get_llm(name, temperature=None, top_p=None) for name in LLM_NAMES]


def is_function_calling(llm: LLM):
    try:
        if getattr(llm.metadata, "is_function_calling_model", False):
            if "flash" in llm.metadata.model_name:
                return False
            return True
    except ValueError:
        return False


def get_nothink_str(llm: LLM):
    if "glm-" in llm.metadata.model_name.lower():
        return "/nothink"
    return "/no_think"


def get_tokenizer(
    name: str,
) -> T.Callable[
    [
        str,
        DefaultNamedArg(T.Literal["all"] | T.AbstractSet[str], "allowed_special"),  # type: ignore
        DefaultNamedArg(T.Literal["all"] | T.Collection[str], "disallowed_special"),  # type: ignore
    ],
    list[int],
]:
    return tiktoken.encoding_for_model("gpt-4o-mini").encode  # type: ignore



from syftr.configuration import cfg
from syftr.hf_endpoint_embeddings import HFEndpointEmbeddings


def get_hf_endpoint_embed_model(model_name: str) -> HFEndpointEmbeddings | None:
    embed_model_cfg = cfg.hf_embeddings.models_config_map.get(model_name, None)
    if not embed_model_cfg:
        raise ValueError(f"Model {model_name} not found in config")
    embed_model = HFEndpointEmbeddings(
        hf_api_key=str(cfg.hf_embeddings.api_key.get_secret_value()),
        hf_api_url=embed_model_cfg.api_url,
        model_name=embed_model_cfg.embedding_model_name,
        max_length=embed_model_cfg.max_length,
        hf_embedding_batch_size=embed_model_cfg.hf_embedding_batch_size,
        query_prefix=embed_model_cfg.query_prefix,
        text_prefix=embed_model_cfg.text_prefix,
    )

    return embed_model



import typing as T

import pytest
from llama_index.core.callbacks import CBEventType

from syftr.flows import CritiqueAgentFlow, ReActAgentFlow, SubQuestionRAGFlow
from syftr.logger import logger
from syftr.studies import StudyConfig

QA_PAIRS = {
    "hotpotqa_hf/dev": [
        ("Were Scott Derrickson and Ed Wood both Americans?", "yes"),
    ],
    "hotpot/dev": [
        ("Were Scott Derrickson and Ed Wood both Americans?", "yes"),
    ],
    "crag/music": [
        ("what is the song that dua lipa did with elton john?", "cold heart")
    ],
    "financebench": [
        (
            "As of FY2023Q1, why did Pepsico raise full year guidance for FY2023?",
            "Pepsico experienced a strong start to FY2023.",
        ),
    ],
}


@pytest.mark.xfail(reason="Tool use not guaranteed for all models")
def test_critique_agent_flow(
    critique_agent_flow: T.Tuple[CritiqueAgentFlow, StudyConfig],
    llama_debug,
):
    flow, study_config = critique_agent_flow
    for question, _ in QA_PAIRS[study_config.dataset.name]:
        _, _, call_data = flow.generate(question)
        assert call_data
        assert llama_debug.get_event_pairs(CBEventType.LLM)
        assert llama_debug.get_event_pairs(CBEventType.AGENT_STEP)
        if not llama_debug.get_event_pairs(CBEventType.SUB_QUESTION):
            logger.warning("Critique agent did not use sub-questions.")
        assert llama_debug.get_event_pairs(CBEventType.SYNTHESIZE)


@pytest.mark.xfail(reason="Some models hit the max iterations limit")
def test_react_flow(
    react_agent_flow: T.Tuple[ReActAgentFlow, StudyConfig],
    llama_debug,
):
    flow, study_config = react_agent_flow
    for question, _ in QA_PAIRS[study_config.dataset.name]:
        _, _, call_data = flow.generate(question)  # type: ignore
        assert call_data
        assert llama_debug.get_event_pairs(CBEventType.LLM)
        assert llama_debug.get_event_pairs(CBEventType.AGENT_STEP)
        assert llama_debug.get_event_pairs(CBEventType.FUNCTION_CALL)
        assert llama_debug.get_event_pairs(CBEventType.SYNTHESIZE)


@pytest.mark.flaky(reruns=4, reruns_delay=2)
def test_subquestion_flow(
    real_subquestion_flow: T.Tuple[SubQuestionRAGFlow, StudyConfig],
    llama_debug,
):
    flow, study_config = real_subquestion_flow
    for question, _ in QA_PAIRS[study_config.dataset.name]:
        _, _, call_data = flow.generate(question)  # type: ignore
        assert len(call_data) > 1, call_data
        assert sum(call.cost for call in call_data) > 0, call_data
        assert sum(call.input_tokens for call in call_data) > 0, call_data
        assert sum(call.output_tokens for call in call_data) > 0, call_data
        assert llama_debug.get_event_pairs(CBEventType.LLM)
        assert llama_debug.get_event_pairs(CBEventType.SUB_QUESTION)


@pytest.mark.flaky(reruns=4, reruns_delay=2)
def test_react_agent_flow_hybrid_hyde_reranker_few_shot(
    react_agent_flow_hybrid_hyde_reranker_few_shot, llama_debug
):
    flow, study_config = react_agent_flow_hybrid_hyde_reranker_few_shot
    for question, _ in QA_PAIRS[study_config.dataset.name]:
        flow.generate(question)
    assert llama_debug.get_event_pairs(CBEventType.AGENT_STEP)


@pytest.mark.flaky(reruns=4, reruns_delay=2)
def test_coa_agent_flow(coa_agent_flow, llama_debug):
    flow, study_config = coa_agent_flow
    for question, _ in QA_PAIRS[study_config.dataset.name]:
        _, _, call_data = flow.generate(question)
        assert call_data
        assert llama_debug.get_event_pairs(CBEventType.LLM)
        assert llama_debug.get_event_pairs(CBEventType.QUERY)
        assert llama_debug.get_event_pairs(CBEventType.AGENT_STEP)
        assert llama_debug.get_event_pairs(CBEventType.SYNTHESIZE)


def test_coa_agent_flow_math(coa_agent_flow, llama_debug):
    flow, study_config = coa_agent_flow
    response, _, _ = flow.generate(
        "what is 123.123*101.101 and what is its product with 12345. "
        "then what is 415.151 - 128.24 and what is its product with the previous product?"
    )
    answer = (123.123 * 101.101) * 12345 * (415.151 - 128.24)
    valid_answers = [str(answer), f"{answer:,}", f"{answer:,}".replace(",", "\u202f")]
    assert any(answer in response.text for answer in valid_answers), (
        f"{valid_answers=}, {response.text=}"
    )



from syftr.helpers import get_exception_report


def test_get_exception_report():
    generation_exceptions = []

    try:
        eval("x === 3")  # SyntaxError
    except Exception as e:
        generation_exceptions.append(e)

    try:
        "2" + 2  # TypeError
    except Exception as e:
        generation_exceptions.append(e)

    try:
        int("abc")  # ValueError
    except Exception as e:
        generation_exceptions.append(e)

    evaluation_exceptions = []

    try:
        lst = [1, 2, 3]
        print(lst[5])  # IndexError
    except Exception as e:
        evaluation_exceptions.append(e)

    try:
        d = {"a": 1}
        print(d["b"])  # KeyError
    except Exception as e:
        evaluation_exceptions.append(e)

    try:
        None.some_method()  # AttributeError
    except Exception as e:
        evaluation_exceptions.append(e)

    try:
        1 / 0  # ZeroDivisionError
    except Exception as e:
        evaluation_exceptions.append(e)

    exceptions = []
    exceptions.append(
        ExceptionGroup(
            "Exceptions during generation",
            generation_exceptions,
        )
    )
    exceptions.append(
        ExceptionGroup(
            "Exceptions during evaluation",
            evaluation_exceptions,
        )
    )
    exception_group = ExceptionGroup("Trial failed", exceptions)
    exception_message = get_exception_report(exception_group)

    assert "SyntaxError" in exception_message
    assert "TypeError" in exception_message
    assert "ValueError" in exception_message
    assert "IndexError" in exception_message
    assert "KeyError" in exception_message
    assert "AttributeError" in exception_message
    assert "ZeroDivisionError" in exception_message
    assert "Trial failed" in exception_message
    assert "Exceptions during generation" in exception_message
    assert "Exceptions during evaluation" in exception_message



import inspect
import typing as T
from typing import Iterator

import optuna
import pytest

import syftr.baselines as baselines
from syftr.llm import LLM_NAMES, get_tokenizer
from syftr.studies import ParamDict, SearchSpace


@pytest.fixture(scope="session")
def optuna_study() -> optuna.Study:
    optuna.logging.set_verbosity(optuna.logging.WARNING)
    study = optuna.study.create_study(
        directions=[
            "maximize",
            "minimize",
        ],
        sampler=optuna.samplers.TPESampler(
            multivariate=True,
            group=True,
        ),
    )
    return study


@pytest.fixture(params=[f"random_trial_{i}" for i in range(100)])
def search_space_and_trial_params(
    request, optuna_study
) -> T.Tuple[SearchSpace, T.Dict[str, T.Any]]:
    ss = SearchSpace()
    distributions = ss.build_distributions()
    trial = optuna_study.ask(distributions)
    params = ss.sample(trial)
    assert params
    return ss, params


@pytest.fixture(params=LLM_NAMES)
def tokenizer(request):
    llm_name = request.param
    return get_tokenizer(llm_name)


def baseline_templates() -> Iterator[ParamDict]:
    for name, value in inspect.getmembers(baselines):
        if name.endswith("_TEMPLATE"):
            yield value
    for baseline in baselines.INDIVIDUAL_BASELINES:
        yield baseline


@pytest.fixture(params=baseline_templates())
def baseline_template(request):
    return request.param



from pathlib import Path
from typing import Dict, List, Optional

from bs4 import BeautifulSoup
from fsspec import AbstractFileSystem
from html2text import html2text
from llama_index.core.readers.base import BaseReader
from llama_index.core.readers.file.base import get_default_fs
from llama_index.core.schema import Document


class HTMLFileMarkdownReader(BaseReader):
    """Reads HTML files into markdown using html2text package."""

    def load_data(
        self,
        file: Path,
        extra_info: Optional[Dict] = None,
        fs: Optional[AbstractFileSystem] = None,
    ) -> List[Document]:
        fs = fs or get_default_fs()

        with fs.open(file, "rb") as fp:
            soup = BeautifulSoup(fp.read(), "html.parser")
            title = next(iter(soup.find_all("title")), None)
            metadata = {
                "file_name": file.name[:256],
                "page_title": str(title)[:256],
            }
            if extra_info:
                metadata.update(extra_info)
        with fs.open(file, "r") as fp:
            text = fp.read()
            text = html2text(text)
        return [Document(text=text, extra_info=metadata)]



import getpass
import typing as T
from copy import deepcopy
from pathlib import Path

import yaml

from syftr.configuration import cfg
from syftr.storage import DRDocsHF, SyftrQADataset
from syftr.studies import (
    CritiqueRAGAgent,
    Evaluation,
    FewShotRetriever,
    Hyde,
    LATSRagAgent,
    LLMConfig,
    OptimizationConfig,
    QueryDecomposition,
    ReactRAGAgent,
    Reranker,
    Retriever,
    SearchSpace,
    Splitter,
    StudyConfig,
    SubQuestionRAGAgent,
    TimeoutConfig,
    TopK,
    TransferLearningConfig,
)


def _build_study_name(
    dataset: SyftrQADataset,
    bench_num: int,
    prefix: str,
    run_name: str,
    add_username: bool | None = False,
) -> str:
    dataset_name = dataset.name.replace("/", "-")
    study_name = f"{prefix}{bench_num}--{run_name}--{dataset_name}"
    if hasattr(dataset, "subset") and dataset.subset:
        study_name += f"--{dataset.subset}"
    if add_username:
        username = getpass.getuser()
        study_name = f"{username}--" + study_name
    return study_name


def build_configs(
    datasets: T.List[SyftrQADataset],
    search_space: SearchSpace,
    optimization_config: OptimizationConfig,
    evaluation: Evaluation,
    bench_num: int,
    reuse_study: bool,
    recreate_study: bool,
    prefix: str,
    run_name: str,
    embedding_max_time: int,
    transfer_learning: TransferLearningConfig | None = None,
    add_username: bool | None = False,
    is_private: bool | None = False,
) -> T.Tuple[T.List[StudyConfig], T.List[Path]]:
    configs = []
    for dataset in datasets:
        _search_space = deepcopy(search_space)
        configs.append(
            StudyConfig(
                name=_build_study_name(
                    dataset,
                    bench_num,
                    prefix,
                    run_name,
                    add_username,
                ),
                dataset=dataset,
                search_space=_search_space,
                optimization=optimization_config,
                # transfer_learning=transfer_learning,
                toy_mode=False,
                reuse_study=reuse_study,
                recreate_study=recreate_study,
                evaluation=evaluation,
                timeouts=TimeoutConfig(
                    embedding_timeout_active=True, embedding_max_time=embedding_max_time
                ),
            )
        )
    paths = []
    for config in configs:
        file_name = f"{config.name}.yaml"
        if is_private:
            file_name = f"private--{file_name}"
        paths.append(cfg.paths.studies_dir / file_name)
    for config, path in zip(configs, paths):
        with open(path, "w") as handle:
            yaml.dump(config.dict(), handle)

    paths = [path.relative_to(cfg.paths.root_dir) for path in paths]

    return configs, paths


def build_example_config(
    llms: T.List[str],
    embedding_models: T.List[str],
    num_trials: int | None = 10,
    max_concurrent_trials: int | None = 1,
    num_eval_samples: int | None = 50,
    num_random_trials: int | None = 10,
    embedding_max_time: int | None = 3600 * 8,
    datasets: T.List[SyftrQADataset] | None = None,
    optimization_config: OptimizationConfig | None = None,
    reuse_study: bool | None = True,
    recreate_study: bool | None = False,
    add_username: bool | None = False,
):
    assert isinstance(num_trials, int), "num_trials must be an int"
    assert isinstance(max_concurrent_trials, int), (
        "max_concurrent_trials must be an int"
    )
    assert isinstance(num_eval_samples, int), "num_eval_samples must be an int"
    assert isinstance(num_random_trials, int), "num_random_trials must be an int"
    assert isinstance(embedding_max_time, int), "embedding_max_time must be an int"
    assert reuse_study is not None, "reuse_study must be a boolean"
    assert recreate_study is not None, "recreate_study must be a boolean"
    assert add_username is not None, "add_username must be a boolean"

    datasets = datasets or [DRDocsHF()]

    optimization_config = optimization_config or OptimizationConfig(
        method="expanding",
        shuffle_blocks=False,
        num_trials=num_trials,
        baselines_cycle_llms=False,
        shuffle_baselines=True,
        max_concurrent_trials=max_concurrent_trials,
        num_eval_samples=num_eval_samples,
        num_eval_batch=5,
        rate_limiter_max_coros=10,
        rate_limiter_period=60,
        max_trial_cost=40.0,
        cpus_per_trial=1,
        seeder_timeout=3600 * 1,  # None: wait until finished, 0: don't wait
        num_random_trials=num_random_trials,
        use_individual_baselines=False,
        use_agent_baselines=False,
        use_variations_of_baselines=False,
        use_pareto_baselines=False,
        use_pareto_pruner=True,
        use_cost_pruner=True,
        use_runtime_pruner=True,
        use_toy_baselines=False,
        sampler="tpe",
    )

    search_space = SearchSpace(
        few_shot_enabled=[False, True],
        additional_context_enabled=[False, True],
        hyde_enabled=[False, True],
        reranker_enabled=[False, True],
        splitter=Splitter(
            methods=[
                "recursive",
                "sentence",
                "token",
            ],
            chunk_min_exp=7,
            chunk_max_exp=10,
            chunk_overlap_frac_min=0.0,
            chunk_overlap_frac_max=0.5,
            chunk_overlap_frac_step=0.05,
        ),
        rag_modes=[
            "no_rag",
            "rag",
            "lats_rag_agent",
            "react_rag_agent",
            "critique_rag_agent",
            "sub_question_rag",
        ],
        template_names=[
            "default",
            "concise",
            "CoT",
            # "finance-expert",
        ],
        response_synthesizer_llm_config=LLMConfig(llm_names=llms),
        rag_retriever=Retriever(
            embedding_models=embedding_models,
            methods=["dense", "sparse", "hybrid"],
            top_k=TopK(kmin=1, kmax=10, log=False),
            query_decomposition=QueryDecomposition(
                llm_config=LLMConfig(llm_names=llms),
                num_queries_min=2,
                num_queries_max=5,
                num_queries_step=1,
            ),
        ),
        react_rag_agent=ReactRAGAgent(
            subquestion_engine_llm_config=LLMConfig(llm_names=llms),
            subquestion_response_synthesizer_llm_config=LLMConfig(llm_names=llms),
        ),
        sub_question_rag=SubQuestionRAGAgent(
            subquestion_engine_llm_config=LLMConfig(llm_names=llms),
            subquestion_response_synthesizer_llm_config=LLMConfig(llm_names=llms),
        ),
        critique_rag_agent=CritiqueRAGAgent(
            subquestion_engine_llm_config=LLMConfig(llm_names=llms),
            subquestion_response_synthesizer_llm_config=LLMConfig(llm_names=llms),
            critique_agent_llm_config=LLMConfig(llm_names=llms),
            reflection_agent_llm_config=LLMConfig(llm_names=llms),
        ),
        lats_rag_agent=LATSRagAgent(),
        reranker=Reranker(llm_config=LLMConfig(llm_names=llms)),
        hyde=Hyde(llm_config=LLMConfig(llm_names=llms)),
        few_shot_retriever=FewShotRetriever(
            embedding_models=embedding_models,
        ),
    )

    evaluation = Evaluation(
        mode="random",
        raise_on_exception=False,
    )
    evaluation.llm_names = llms

    configs, paths = build_configs(
        datasets=datasets,
        search_space=search_space,
        optimization_config=optimization_config,
        evaluation=evaluation,
        bench_num=1,
        reuse_study=reuse_study,
        recreate_study=recreate_study,
        prefix="example",
        run_name="small-study",
        embedding_max_time=embedding_max_time,
        transfer_learning=None,
        add_username=add_username,
    )
    return configs, paths



from syftr.flows import Flow


def test_loaded_models(configured_models):
    name, llm = configured_models
    llm.complete("hello world")


def test_loaded_models_in_flow(configured_models):
    name, llm = configured_models
    flow = Flow(response_synthesizer_llm=llm)
    response, duration, data = flow.generate("hello world")
    assert response.text
    assert duration
    assert data[0].input_tokens
    assert data[0].output_tokens



import random
import re
import string
import typing as T

from pydantic import BaseModel, Field

from syftr.configuration import NDIGITS


def normalize_text(s):
    def remove_articles(text):
        regex = re.compile(r"\b(a|an|the)\b", re.UNICODE)
        return re.sub(regex, " ", text)

    def white_space_fix(text):
        return " ".join(text.split())

    def remove_punc(text):
        exclude = set(string.punctuation)
        return "".join(ch for ch in text if ch not in exclude)

    def lower(text):
        return text.lower()

    return white_space_fix(remove_articles(remove_punc(lower(s))))


def f1_score(prediction, truth):
    pred_tokens = normalize_text(prediction).split()
    truth_tokens = normalize_text(truth).split()

    # if either the prediction or the truth is no-answer then f1 = 1 if they agree, 0 otherwise
    if len(pred_tokens) == 0 or len(truth_tokens) == 0:
        return int(pred_tokens == truth_tokens)

    common_tokens = set(pred_tokens) & set(truth_tokens)

    # if there are no common tokens then f1 = 0
    if len(common_tokens) == 0:
        return 0
    prec = len(common_tokens) / len(pred_tokens)
    rec = len(common_tokens) / len(truth_tokens)

    return 2 * (prec * rec) / (prec + rec)


class QAPair(BaseModel):
    question: str
    answer: str
    id: str
    context: T.Union[T.Dict[str, str], T.List[T.Dict]]
    supporting_facts: T.List[T.Any]
    difficulty: str
    qtype: str
    gold_evidence: T.List[str] = Field(
        default_factory=list,
        description="List of gold text snippets that must be retrieved exactly for full recall",
    )


class QARTriplet(QAPair):
    response: str
    label: bool


class RandomTrial:
    """A dummy trial class for generating random parameters."""

    def suggest_categorical(self, name, choices):
        return random.choice(choices)

    def suggest_int(self, name, low, high, step=1, log=False):
        return random.randrange(low, high + 1, step)

    def suggest_float(self, name, low, high, step=None, log=False):
        if step:
            num_steps = int((high - low) / step)
            return low + step * random.randint(0, num_steps)
        value = random.uniform(low, high)
        return round(value, NDIGITS)



import asyncio
import json
from pathlib import Path
from typing import List

import click
import pandas as pd
from aiolimiter import AsyncLimiter

from syftr.core import RandomTrial
from syftr.evaluation.evaluation import (
    ExactMatchEvaluator,
    SyftrEvaluationResult,
    _aeval_retriever_pair,
)
from syftr.flows import RetrieverFlow
from syftr.instrumentation.arize import instrument_arize
from syftr.llm import get_llm
from syftr.retrievers.build import build_dummy_retriever
from syftr.studies import RetrieverStudyConfig
from syftr.tuner.qa_tuner import build_flow


@click.command()
@click.option(
    "-s",
    "--study-config-path",
    type=click.Path(exists=True, dir_okay=False, path_type=Path),
    required=True,
    help="Path to the study config yaml file",
)
@click.option("-d", "--dummy", is_flag=True, help="Use dummy retriever flow")
@click.option("-i", "--interactive", is_flag=True)
@click.option("-o", "--instrumentation", is_flag=True)
@click.option("-f", "--flow-json", type=str, help="JSON string of flow parameters")
def main(
    study_config_path: Path,
    instrumentation: bool,
    interactive: bool,
    flow_json: str,
    dummy: bool,
):
    study_config: RetrieverStudyConfig = RetrieverStudyConfig.from_file(
        study_config_path
    )
    click.secho(
        f"Loaded study config for study {study_config.name}", bold=True, fg="green"
    )

    if instrumentation:
        instrument_arize()

    if dummy:
        click.secho("Building dummy retriever flow", fg="yellow")
        retriever, docstore = build_dummy_retriever(study_config)
        flow = RetrieverFlow(
            response_synthesizer_llm=get_llm("gpt-4o-mini"),
            retriever=retriever,
            docstore=docstore,
            hyde_llm=None,
            additional_context_num_nodes=0,
            params={},
            enforce_full_evaluation=False,
        )
        click.secho("Built dummy retriever flow", fg="green")
    else:
        if flow_json:
            click.secho(f"Building flow from JSON: '{flow_json}'", fg="yellow")
            flow_params = json.loads(flow_json)

        else:
            search_space = study_config.search_space
            random_trial = RandomTrial()
            flow_params = search_space.sample(random_trial)
            flow_json = json.dumps(flow_params, ensure_ascii=False)
            click.secho(f"Building random flow '{flow_json}'", fg="yellow")

        f = build_flow(flow_params, study_config)
        assert isinstance(f, RetrieverFlow), "Flow must be a RetrieverFlow"
        flow = f
        click.secho(f"Built flow '{flow_json}'", fg="green")

    click.secho(f"Loading QA Pairs from {study_config.dataset.name}", fg="yellow")
    dataset = list(study_config.dataset.iter_examples())
    filtered_dataset = [pair for pair in dataset if pair.gold_evidence]
    if not filtered_dataset:
        raise ValueError("No QAPairs with gold_evidence found in the dataset.")
    click.secho(f"Loaded {len(filtered_dataset)} QA Pairs", fg="green")

    rate_limiter = AsyncLimiter(
        study_config.optimization.rate_limiter_max_coros,
        study_config.optimization.rate_limiter_period,
    )
    evaluator = ExactMatchEvaluator()

    results = []
    total_recall = 0.0
    total_context_length = 0.0
    for i, pair in enumerate(filtered_dataset):
        click.secho("Question: ", fg="blue", nl=False)
        click.echo(f"{pair.question}")
        click.secho("Reference Answer: ", fg="blue", nl=False)
        click.echo(f"{pair.answer}\n")
        if interactive:
            breakpoint()
        result = asyncio.run(
            _aeval_retriever_pair(
                pair,
                flow,
                [evaluator],
                rate_limiter,
                raise_on_exception=study_config.evaluation.raise_on_exception,
            )
        )
        recall = result.retriever_recall
        context_length = result.retriever_context_length
        passing = result.passing
        if result.generation_exception:
            click.secho(
                f"Generation Exception: {result.generation_exception}", fg="red"
            )
        if result.evaluation_exception:
            click.secho(
                f"Evaluation Exception: {result.evaluation_exception}", fg="red"
            )
        click.secho(
            f"Recall: {recall} with context length {context_length}",
            fg="green" if recall and recall >= 1.0 else "red",
        )
        click.secho(
            f"Evaluation Result: {'PASS' if passing else 'FAIL'}: {result.score}/5",
            fg="green" if passing else "red",
        )
        results.append(result)
        total_recall += recall if recall is not None else 0.0
        total_context_length += context_length if context_length is not None else 0.0

    avg_recall = total_recall / len(filtered_dataset)
    avg_context_length = total_context_length / len(filtered_dataset)
    click.echo(
        f"Average Recall: {avg_recall} Avg Context: {avg_context_length} in {len(filtered_dataset)} QA pairs"
    )


def write_results(results: List[SyftrEvaluationResult]):
    data = [result.model_dump() for result in results]
    df = pd.DataFrame(data)
    df.to_csv("flow-results.csv")
    return df


if __name__ == "__main__":
    main()



import functools
import json
import time
import typing as T
from typing import Protocol

from llama_index.core import PromptTemplate, VectorStoreIndex
from llama_index.core.agent import ReActAgent
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.postprocessor import LLMRerank
from llama_index.core.postprocessor.types import BaseNodePostprocessor
from llama_index.core.prompts import PromptType
from llama_index.core.query_engine import RetrieverQueryEngine, TransformQueryEngine
from llama_index.core.tools import QueryEngineTool, ToolMetadata

from syftr.logger import logger


class AgentFlow:
    """Base class for agentic flows, containing common methods."""

    reranker_top_n: int
    reranker_llm: FunctionCallingLLM | None
    get_examples: T.Callable[[str], str]

    def __init__(
        self,
        llm: FunctionCallingLLM,
        template: str | None = None,
        **kwargs,
    ):
        self.llm = llm
        self.template = template
        for key, value in kwargs.items():
            setattr(self, key, value)
        self.node_postprocessors = []
        if hasattr(self, "reranker_top_n"):
            self.node_postprocessors.append(
                LLMRerank(top_n=self.reranker_top_n, llm=self.reranker_llm)
            )
        self.name = self.__class__.__name__

    def get_prompt(self, query_str: str) -> str:
        if self.template is None:
            return query_str
        if not hasattr(self, "get_examples"):
            return self.template.format(query_str=query_str)
        examples = self.get_examples(query_str)  # type: ignore
        assert examples, "No examples found for few-shot prompting"
        return self.template.format(
            query_str=query_str,
            few_shot_examples=examples,
        )

    def get_prompt_template(self) -> PromptTemplate | None:
        prompt_template = None
        function_mappings = None
        if hasattr(self, "get_examples"):
            function_mappings = {"few_shot_examples": self.get_examples}
        if self.template:
            logger.debug("Creating prompt template from '%s'", self.template)
            prompt_template = PromptTemplate(
                template=self.template,
                prompt_type=PromptType.QUESTION_ANSWER,
                function_mappings=function_mappings,
            )
        return prompt_template


class AgentProtocol(Protocol):
    """Typing protocol for type-hinting Agent flows.

    usage:

        def foo(agent: AgentProtocol):
            agent.generate('hi')
    """

    llm: FunctionCallingLLM
    llm_name: str
    query_engine: RetrieverQueryEngine | TransformQueryEngine
    node_postprocessors: T.List[BaseNodePostprocessor] | None
    agent: ReActAgent
    name: str
    use_hyde: bool
    hyde_llm: FunctionCallingLLM | None

    def get_prompt(self, query_str: str) -> str:
        pass

    def get_prompt_template(self) -> PromptTemplate:
        pass

    def generate(self, query_str: str) -> T.Tuple[str, float]:
        pass

    async def agenerate(self, query_str: str) -> str:
        pass


class AgentMixin:
    """Default agentic generate method implementations.

    Add this to your class inheritance when making a new agent flow:

        class MyAgentFlow(AgentFlow, AgentMixin):
            ...
    """

    def generate(self: AgentProtocol, query_str: str) -> T.Tuple[str, float]:
        start_time = time.perf_counter()
        response = self.agent.chat(query_str)
        duration = time.perf_counter() - start_time
        return response, duration

    async def agenerate(self: AgentProtocol, query_str: str) -> T.Tuple[str, float]:
        start_time = time.perf_counter()
        response = await self.agent.achat(query_str)
        duration = time.perf_counter() - start_time
        return response, duration


class LlamaIndexReactRAGAgentFlow(AgentFlow, AgentMixin):
    def __init__(
        self: AgentProtocol,
        indexes: T.List[T.Tuple[str, str, VectorStoreIndex]],
        llm: FunctionCallingLLM,
        system_prompt: PromptTemplate | None = None,
        template: str | None = None,
        **kwargs,
    ):
        super().__init__(llm=llm, template=template, **kwargs)  # type: ignore

        prompt_template: PromptTemplate = self.get_prompt_template()
        query_engines = [
            index.as_query_engine(
                llm=self.llm,
                text_qa_template=prompt_template,
                node_postprocessors=self.node_postprocessors,
            )
            for _, _, index in indexes
        ]
        query_engine_tools = [
            QueryEngineTool(
                query_engine=query_engine,
                metadata=ToolMetadata(
                    name="Grounding_%s" % name,
                    description="Grounding data related to %s" % desc,
                ),
            )
            for (name, desc, _), query_engine in zip(indexes, query_engines)
        ]
        self.agent = ReActAgent.from_tools(
            query_engine_tools,  # type: ignore
            llm=self.llm,
            verbose=True,
        )
        if system_prompt:
            self.agent.update_prompts({"agent_worker:system_prompt": system_prompt})
            self.agent.reset()


class FlowJSONDecodeError(Exception):
    pass


def chain_hasattr(obj, attr):
    for a in attr.split("."):
        if hasattr(obj, a):
            obj = getattr(obj, a)
        else:
            return False
    return True


def chain_getattr(obj, attr):
    return functools.reduce(
        lambda obj, attr: getattr(obj, attr), [obj] + attr.split(".")
    )


class FlowJSONHandler(json.JSONEncoder):
    copy_args: T.List[str] = [
        "fusion_mode",
        "num_queries",
        "retriever_weights",
        "similarity_top_k",
        "use_hyde",
        "use_reranker",
        "reranker_top_n",
        "retriever",
        "embedding_model",
        "splitter",
        "template_name",
        "rag_mode",
        "llm_name",
        "reranker_llm_name",
        "hyde_llm_name",
        "chunk_size",
        "chunk_overlap",
    ]

    def default(self, obj):
        out = {}
        for arg in self.copy_args:
            if hasattr(obj, arg):
                out[arg] = getattr(obj, arg)
        retrievers = []
        if chain_hasattr(obj, "query_engine.retriever"):
            retrievers.append({"name": obj.query_engine.retriever.__class__.__name__})
        elif chain_hasattr(obj, "query_engine._query_engine.retriever"):
            retrievers.append(
                {"name": obj.query_engine._query_engine.retriever.__class__.__name__}
            )
        if retrievers:
            out["retrievers"] = retrievers
        if hasattr(obj, "retrievers"):
            retrievers += obj.retrievers
            out["retrievers"] = retrievers
        if hasattr(obj, "agent"):
            out["agent"] = obj.agent.__class__.__name__
        return out



import logging
import queue
import threading
from contextlib import contextmanager
from pathlib import Path

from rich.console import Console
from rich.text import Text
from sqlalchemy.exc import OperationalError
from transformers import AutoConfig

from syftr import __version__
from syftr.configuration import SYFTR_CONFIG_FILE_ENV_NAME, cfg
from syftr.llm import LLM_NAMES, get_llm
from syftr.optuna_helper import get_study_names
from syftr.studies import DEFAULT_EMBEDDING_MODELS
from syftr.studyconfig_helper import build_example_config

llms = []
embedding_models = []

console = Console()


def print_into():
    ascii_art = rf"""
Welcome to
 ___  _  _  ____  ____  ____
/ __)( \/ )( ___)(_  _)(  _ \
\__ \ \  /  )__)   )(   )   /
(___/ (__) (__)   (__) (_)\_)

version {__version__}.
Running system check..."""
    # Print the ASCII art as a Text object, ensuring it's treated literally
    # and doesn't get misinterpreted by console markup.
    console.print(Text(ascii_art))
    return True


def check_config():
    assert "yaml_file" in cfg.model_config, "Missing 'yaml_file' in model_config"
    file_locations_str = [
        str(location)
        for location in reversed(cfg.model_config["yaml_file"])  # type: ignore
    ]

    potential_paths = [Path(loc) for loc in file_locations_str if loc != "."]

    existing_config_files = [str(p) for p in potential_paths if p.is_file()]

    if existing_config_files:
        console.print("Syftr will load configuration from the following files:")

        # Display the files that were actually found and will be loaded
        console.print("\n[green]Found:[/green]")
        for f_path in existing_config_files:
            console.print(f"- {f_path}")

        console.print(
            "\nConfiguration values are merged, with files listed earlier taking precedence over later ones."
        )
        # console.print(f"The final effective configuration is available via `cfg`.") # This was in my previous version, user removed it.
        console.print()
        return True

    # If no files were found, guide the user.
    # Show the locations that were checked.
    # To match user's logic, we base this on the file_locations_str
    checked_locations_display = [
        str(Path(loc).absolute()) for loc in file_locations_str if loc != "."
    ]

    console.print(
        "[yellow]No syftr configuration files (e.g., config.yaml, .syftr.yaml) were found.[/yellow]"
    )
    console.print("Please create a configuration file in one of these locations:")

    for loc in checked_locations_display:
        console.print(f"- {loc}")

    console.print(f"""
or specify its path using the environment variable {SYFTR_CONFIG_FILE_ENV_NAME}.
The README.md file contains an example config.yaml file.
""")
    return False


def check_database():
    db_connections = []
    # Ensure dsn is not None and has hosts method
    if cfg.database and "sqlite" in cfg.database.dsn.unicode_string():
        if not cfg.ray.local:
            console.print(
                "Found default configuration of SQLite. SQLite cannot be used in non-local mode."
            )
            console.print(
                "Set ray.local = True or configure PostgreSQL for advanced usage."
            )
            return False
        console.print("Found default configuration of SQLite. Should work locally...")
        console.print("For advanced usage, consider configuring PostgreSQL.")
        return True
    if (
        cfg.database
        and cfg.database.dsn
        and hasattr(cfg.database.dsn, "hosts")
        and callable(cfg.database.dsn.hosts)  # type: ignore
    ):
        try:
            # hosts() might return a list of dicts or a list of pydantic models
            parsed_hosts = cfg.database.dsn.hosts()  # type: ignore
            if parsed_hosts:
                for host_info in parsed_hosts:  # type: ignore
                    if isinstance(host_info, dict):
                        host = host_info.get("host")
                        port = host_info.get("port")
                    elif hasattr(host_info, "host") and hasattr(
                        host_info, "port"
                    ):  # Pydantic model like
                        host = host_info.host
                        port = host_info.port
                    else:
                        db_connections.append(f"Unknown host structure: {host_info}")
                        continue

                    if host and port:
                        db_connections.append(f"{host}:{port}")
                    elif host:
                        db_connections.append(f"{host} (default port)")
                    else:
                        db_connections.append("Invalid host entry")
            else:  # If hosts() returns empty or None
                db_connections.append(f"No host information in DSN: {cfg.database.dsn}")

        except Exception as e:
            db_connections.append(
                f"Could not parse DSN: {cfg.database.dsn}. Error: {e}"
            )
    elif cfg.database and cfg.database.dsn:
        db_connections.append(f"Attempting direct DSN: {cfg.database.dsn}")
    else:
        db_connections.append("Postgres DSN not configured.")

    console.print("Checking connection to database(s) based on DSN:")
    if db_connections:
        for conn_info in db_connections:
            console.print(f"- {conn_info}")
    else:
        console.print(
            "- [yellow]No database connection details found in configuration.[/yellow]"
        )
    console.print()

    try:
        study_names = get_study_names(".*")
        console.print(
            f"[green]Database connection successful. We found {len(study_names)} studies.[/green]\n"
        )
    except OperationalError as e:
        console.print("[bold red]Postgres database connection failed.[/bold red]")
        console.print(f"[bold red]Error details: {e}[/bold red]")
        console.print(
            "[yellow]Please check your database settings and configuration.[/yellow]"
        )
        console.print(
            Text(
                """
Once you have installed PostgreSQL, you can do the following setup from a Linux bash:

  sudo -u postgres psql
  CREATE USER syftr WITH PASSWORD 'your_password';
  CREATE DATABASE syftr WITH OWNER syftr;
  \\q

In your config.yaml file, ensure the 'postgres.dsn' is correctly set, for example:

postgres:
  dsn: "postgresql://syftr:your_password@localhost:5432/syftr"

You may need to adjust the username, password, hostname, and port depending on your setup.
Make sure the PostgreSQL server is running and accessible from where you are running this script.
""",
                style="yellow",
            )
        )
        return False
    return True


def _check_single_llm_worker(llm_name: str, results_queue: queue.Queue):
    """
    Worker function for a thread to check a single LLM instance.
    Puts a tuple (llm_name, status, detail_string) into the results_queue.
    status can be "accessible", "inaccessible", "warning".
    """
    try:
        llm_instance = get_llm(llm_name)

        # Perform a simple synchronous completion test.
        response = llm_instance.complete(
            "Just respond with the phrase `[OK]' /no_think"
        )

        if response and hasattr(response, "text") and response.text:
            response_snippet = (
                response.text[:70].replace("\n", " ") + "..."
                if len(response.text) > 70
                else response.text.replace("\n", " ")
            )
            results_queue.put(
                (llm_name, "accessible", f'Responded: "{response_snippet}"')
            )
        else:
            results_queue.put(
                (
                    llm_name,
                    "warning",
                    "Connected but received an empty or unexpected response.",
                )
            )
    except Exception as e:
        error_type = type(e).__name__
        error_message = str(e).replace("\n", " ")
        error_snippet = (
            error_message[:200] + "..." if len(error_message) > 200 else error_message
        )
        results_queue.put((llm_name, "inaccessible", f"{error_type}: {error_snippet}"))


def check_llms():
    """
    Checks accessibility of all LLMs defined in ALL_LLMS using threads.
    Reports a summary after all checks are complete.
    This function is synchronous and integrates with the existing synchronous check workflow.
    """
    global llms  # noqa: PLW0602, PLW0603

    console.print("Checking configured Large Language Models (LLMs)...")

    console.print(
        f"Preparing to check {len(LLM_NAMES)} LLM(s) concurrently using threads: {', '.join(LLM_NAMES)}"
    )

    results_queue = queue.Queue()
    threads = []

    for llm_name in LLM_NAMES:
        thread = threading.Thread(
            target=_check_single_llm_worker, args=(llm_name, results_queue)
        )
        threads.append(thread)
        thread.start()

    console.print(
        f"Launched {len(threads)} threads for LLM checks. Waiting for completion..."
    )

    for thread in threads:
        thread.join()

    console.print("All LLM check threads completed.")

    # Collect results
    results_data = []
    while not results_queue.empty():
        results_data.append(results_queue.get())

    accessible_llms = []
    inaccessible_llms = []
    warning_llms = []

    for result_item in results_data:
        if isinstance(result_item, tuple) and len(result_item) == 3:
            name, status, detail = result_item
            if status == "accessible":
                accessible_llms.append({"name": name, "detail": detail})
            elif status == "inaccessible":
                inaccessible_llms.append({"name": name, "detail": detail})
            elif status == "warning":
                warning_llms.append({"name": name, "detail": detail})
        else:
            console.print(
                f"[bold red]Unexpected result format from LLM check queue:[/bold red] {result_item}"
            )
            inaccessible_llms.append(
                {
                    "name": "Unknown LLM (processing error)",
                    "detail": f"Unexpected result: {result_item}",
                }
            )

    # Print Summary Report
    console.print("\n[bold underline]LLM Accessibility Report[/bold underline]")
    console.print(f"Total LLMs configured and checked: {len(LLM_NAMES)}")

    if accessible_llms:
        console.print(f"\n[green]Accessible LLMs ({len(accessible_llms)}):[/green]")
        for llm_info in accessible_llms:
            console.print(
                f"  [+] [cyan]{llm_info['name']:<25}[/cyan] - {llm_info['detail']}"
            )

    if warning_llms:
        console.print(f"\n[yellow]LLMs with Warnings ({len(warning_llms)}):[/yellow]")
        for llm_info in warning_llms:
            console.print(
                f"  [!] [cyan]{llm_info['name']:<25}[/cyan] - {llm_info['detail']}"
            )

    if inaccessible_llms:
        console.print(
            f"\n[bold red]Inaccessible LLMs ({len(inaccessible_llms)}):[/bold red]"
        )
        for llm_info in inaccessible_llms:
            console.print(
                f"  [-] [cyan]{llm_info['name']:<25}[/cyan] - {llm_info['detail']}"
            )

    console.print("-" * 60)

    llms = [a["name"] for a in accessible_llms]

    if not accessible_llms:
        return False
    return True


@contextmanager
def suppress_logging(level=logging.WARNING):
    logger = logging.getLogger("syftr")
    old_level = logger.level
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(old_level)


def _check_single_embedding_worker(model_name: str, results_queue: queue.Queue):
    """
    Worker function for a thread to check a single embedding model instance.
    Puts a tuple (model_name, status, detail_string) into the results_queue.
    status can be "accessible" or "inaccessible".
    """
    try:
        with suppress_logging():  # suppress INFO and DEBUG logs
            AutoConfig.from_pretrained(model_name)
        results_queue.put(
            (
                model_name,
                "accessible",
                "available on HuggingFace.",
            )
        )
    except Exception as e:
        error_type = type(e).__name__
        error_message = str(e).replace("\n", " ")
        error_snippet = (
            error_message[:200] + "..." if len(error_message) > 200 else error_message
        )
        results_queue.put(
            (model_name, "inaccessible", f"{error_type}: {error_snippet}")
        )


def check_embedding_models():
    global embedding_models  # noqa: PLW0602, PLW0603

    console.print("Checking configured Embedding Models...")
    console.print(
        f"Preparing to check {len(DEFAULT_EMBEDDING_MODELS)} model(s) concurrently using threads: {', '.join(DEFAULT_EMBEDDING_MODELS)}"
    )

    results_queue = queue.Queue()
    threads = []

    for model_name in DEFAULT_EMBEDDING_MODELS:
        thread = threading.Thread(
            target=_check_single_embedding_worker, args=(model_name, results_queue)
        )
        threads.append(thread)
        thread.start()

    console.print(
        f"Launched {len(threads)} threads for embedding model checks. Waiting for completion..."
    )

    for thread in threads:
        thread.join()

    console.print("All embedding model check threads completed.")

    results_data = []
    while not results_queue.empty():
        results_data.append(results_queue.get())

    accessible_models = []
    inaccessible_models = []
    warning_models = []

    for result_item in results_data:
        if isinstance(result_item, tuple) and len(result_item) == 3:
            name, status, detail = result_item
            if status == "accessible":
                accessible_models.append({"name": name, "detail": detail})
            elif status == "inaccessible":
                inaccessible_models.append({"name": name, "detail": detail})
            elif status == "warning":
                warning_models.append({"name": name, "detail": detail})
        else:
            console.print(
                f"[bold red]Unexpected result format from embedding model check queue:[/bold red] {result_item}"
            )
            inaccessible_models.append(
                {
                    "name": "Unknown model (processing error)",
                    "detail": f"Unexpected result: {result_item}",
                }
            )

    console.print(
        "\n[bold underline]Embedding Model Accessibility Report[/bold underline]"
    )
    console.print(
        f"Total embedding models configured and checked: {len(DEFAULT_EMBEDDING_MODELS)}"
    )

    if accessible_models:
        console.print(
            f"\n[green]Accessible Embedding Models ({len(accessible_models)}):[/green]"
        )
        for info in accessible_models:
            console.print(f"  [+] [cyan]{info['name']:<25}[/cyan] - {info['detail']}")

    if warning_models:
        console.print(
            f"\n[yellow]Models with Warnings ({len(warning_models)}):[/yellow]"
        )
        for info in warning_models:
            console.print(f"  [!] [cyan]{info['name']:<25}[/cyan] - {info['detail']}")

    if inaccessible_models:
        console.print(
            f"\n[bold red]Inaccessible Embedding Models ({len(inaccessible_models)}):[/bold red]"
        )
        for info in inaccessible_models:
            console.print(f"  [-] [cyan]{info['name']:<25}[/cyan] - {info['detail']}")

    console.print("-" * 60)

    embedding_models = [a["name"] for a in accessible_models]

    if not accessible_models:
        return False
    return True


CHECKS = [
    print_into,
    check_config,
    check_database,
    check_llms,
    check_embedding_models,
]


def check():
    all_passed = True
    for check_func in CHECKS:
        console.rule(f"[bold blue]Running: {check_func.__name__}[/bold blue]")
        if not check_func():
            all_passed = False
            # Specific guidance is printed by the check function itself
            console.print(
                f"[yellow]Check '{check_func.__name__}' failed. Please review the messages above.[/yellow]\n"
            )
        else:
            console.print(f"[green]Check '{check_func.__name__}' passed.[/green]\n")

    console.rule("[bold blue]Summary[/bold blue]")
    if not all_passed:
        console.print("[bold red]One or more checks failed.[/bold red]")
        console.print("""
You can run this script again to check your progress after addressing the issues.
""")
        return False

    console.print("[bold green]All checks passed.[/bold green]")
    console.print("You are good to go!")
    console.print()

    if llms and embedding_models:
        configs, paths = build_example_config(
            llms, embedding_models, reuse_study=False, add_username=True
        )
        console.print(
            "We generated an example configuration based on the available models."
        )
        console.print(
            "By default, the example is running 10 random and 10 optimization trials."
        )
        console.print("You can edit it and then run it from your project root with:")
        console.print()
        console.print(f"[yellow]syftr run {paths[0]}[/yellow]")
        console.print()
        console.print("The configuration as is would (re)create a study with name:")
        console.print()
        console.print(f"[cyan]{configs[0].name}[/cyan]")
        console.print()
    return True


if __name__ == "__main__":
    check()



from pathlib import Path

from syftr.configuration import cfg
from syftr.helpers import get_baselines_from_trials
from syftr.optuna_helper import (
    get_completed_trials,
    get_flows_from_trials,
    get_pareto_flows,
)
from syftr.studies import StudyConfig
from syftr.tuner.qa_tuner import run
from tests.check_trials import check_trials_are_flawless
from tests.e2e.optimization import run_and_test_optimization


def test_reusing_a_study():
    study_config_file_name = "test-one-of-each.yaml"
    study_config_file = Path(cfg.paths.test_studies_dir / study_config_file_name)
    study_config = StudyConfig.from_file(study_config_file)
    study_config.optimization.use_hf_embedding_models = False

    df_trials = run_and_test_optimization(study_config)
    check_trials_are_flawless(df_trials)

    baselines = study_config.optimization.baselines
    baselines_reused = get_baselines_from_trials(df_trials)

    # doing this step by step to get simple error messages
    for b in baselines:
        assert b in baselines_reused, (
            f"Cannot reinstantiate this baseline from flow JSONs: {b}"
        )

    # rerun
    study_config.name = "test-one-of-each-rerun"
    study_config.optimization.baselines = baselines_reused
    df_trials_rerun = run_and_test_optimization(study_config)
    check_trials_are_flawless(df_trials)

    cols = df_trials.columns.str.startswith("params")
    df1 = df_trials.loc[:, cols]

    rerun_rols = df_trials_rerun.columns.str.startswith("params")
    df2 = df_trials_rerun.loc[:, rerun_rols]

    df1_sorted = (
        df1.sort_index(axis=1)
        .sort_values(by=df1.columns.tolist())
        .reset_index(drop=True)
    )
    df2_sorted = (
        df2.sort_index(axis=1)
        .sort_values(by=df2.columns.tolist())
        .reset_index(drop=True)
    )

    assert df1_sorted.equals(df2_sorted)


def test_separate_pareto_evaluation():
    study_config_file_name = "test-one-of-each-pareto.yaml"
    study_config_file = Path(cfg.paths.test_studies_dir / study_config_file_name)
    study_config = StudyConfig.from_file(study_config_file)
    study_config.optimization.use_hf_embedding_models = False

    pareto_flows_test = get_pareto_flows(study_config)

    run(
        study_config=study_config,
        skip_optimization=True,
        skip_pareto=False,
    )
    df_trials = get_completed_trials(study_config.pareto.name)
    pareto_flows_holdout = get_flows_from_trials(df_trials)

    assert sorted(pareto_flows_test, key=lambda d: sorted(d.items())) == sorted(
        pareto_flows_holdout, key=lambda d: sorted(d.items())
    ), "Pareto flows are different"


def test_separate_pareto_replace_llm_evaluation():
    """Tests the evaluation of the Pareto flows but with all LLMs replaced by a single LLM"""
    study_config_file_name = "test-one-of-each-pareto-replace-llm.yaml"
    study_config_file = Path(cfg.paths.test_studies_dir / study_config_file_name)
    study_config = StudyConfig.from_file(study_config_file)
    study_config.optimization.use_hf_embedding_models = False

    pareto_flows_test = get_pareto_flows(study_config)

    run(
        study_config=study_config,
        skip_optimization=True,
        skip_pareto=False,
    )
    df_trials = get_completed_trials(study_config.pareto.name)
    pareto_flows_holdout = get_flows_from_trials(df_trials)

    assert len(pareto_flows_test) == len(pareto_flows_holdout), (
        "Different number of Pareto flows"
    )

    # assert that all flows have the same llm name
    for flow in pareto_flows_holdout:
        llm_name = flow["response_synthesizer_llm_name"]
        assert llm_name == study_config.pareto.replacement_llm_name, (
            "Flow llm name does not match the replacement llm name"
        )



import typing as T

import optuna
from langchain.text_splitter import RecursiveCharacterTextSplitter
from llama_index.core.node_parser import (
    CodeSplitter,
    LangchainNodeParser,
    SentenceSplitter,
    TokenTextSplitter,
)
from llama_index.core.node_parser.interface import NodeParser
from tenacity import retry, stop_after_attempt, wait_fixed
from transformers import AutoTokenizer

from syftr.flows import Flows
from syftr.llm import get_tokenizer
from syftr.optuna_helper import set_metrics
from syftr.studies import StudyConfig


def get_flow_name(rag_mode: str):
    match rag_mode:
        case "no_rag":
            return Flows.GENERATOR_FLOW.value.__name__
        case "rag":
            return Flows.RAG_FLOW.value.__name__
        case "sub_question_rag":
            return Flows.LLAMA_INDEX_SUB_QUESTION_FLOW.value.__name__
        case "react_rag_agent":
            return Flows.LLAMA_INDEX_REACT_AGENT_FLOW.value.__name__
        case "critique_rag_agent":
            return Flows.LLAMA_INDEX_CRITIQUE_AGENT_FLOW.value.__name__
        case "lats_rag_agent":
            return Flows.LLAMA_INDEX_LATS_RAG_AGENT.value.__name__
        case "coa_rag_agent":
            return Flows.LLAMA_INDEX_COA_RAG_AGENT.value.__name__
        case _:
            raise RuntimeError("Cannot identify flow")


def build_splitter(study_config: StudyConfig, params: T.Dict[str, T.Any]) -> NodeParser:
    chunk_size = 2 ** int(params["splitter_chunk_exp"])
    overlap = int(params["splitter_chunk_overlap_frac"] * chunk_size)
    llm_name = params["response_synthesizer_llm_name"]
    match params["splitter_method"]:
        case "html":
            return CodeSplitter(
                language="html",
                max_chars=4 * chunk_size,
            )
        case "sentence":
            return SentenceSplitter(
                chunk_size=chunk_size,
                chunk_overlap=overlap,
                tokenizer=get_tokenizer(llm_name),
            )
        case "token":
            return TokenTextSplitter(
                chunk_size=chunk_size,
                chunk_overlap=overlap,
                tokenizer=get_tokenizer(llm_name),
            )
        case "recursive":
            return LangchainNodeParser(
                RecursiveCharacterTextSplitter.from_huggingface_tokenizer(
                    tokenizer=AutoTokenizer.from_pretrained(
                        params.get("rag_embedding_model", "BAAI/bge-small-en-v1.5")
                    ),
                    chunk_size=chunk_size,
                    chunk_overlap=overlap,
                )
            )
        case _:
            raise ValueError("Invalid splitter")


@retry(stop=stop_after_attempt(6), wait=wait_fixed(10))
def set_trial(
    trial: optuna.trial.FrozenTrial | optuna.trial.Trial,
    study_config: StudyConfig | None = None,
    params: dict[str, str | bool | int | float] | None = None,
    is_seeding: bool | None = None,
    metrics: T.Dict[str, float] | None = None,
    flow_json: str | None = None,
):
    if params:
        rag_mode = params.get("rag_mode")
        if rag_mode is not None:
            flow_name = get_flow_name(str(rag_mode))
            trial.set_user_attr("flow_name", flow_name)
    if study_config:
        trial.set_user_attr("dataset", study_config.dataset.name)
    if is_seeding is not None:
        trial.set_user_attr("is_seeding", is_seeding)
    if flow_json:
        trial.set_user_attr("flow", flow_json)
    if metrics:
        set_metrics(trial, metrics)



from syftr.baselines import set_baselines
from syftr.storage import HotPotQAHF, PartitionMap
from syftr.studies import StudyConfig


def test_set_baselines():
    study_config = StudyConfig(
        name="test-embedding-models",
        dataset=HotPotQAHF(
            partition_map=PartitionMap(test="sample"),
        ),
    )
    study_config.optimization.raise_on_invalid_baseline = True
    set_baselines(study_config)



import argparse
import json
import typing as T

import optuna
import ray

from syftr.configuration import cfg
from syftr.logger import logger
from syftr.optuna_helper import get_pareto_df
from syftr.ray.utils import ray_init
from syftr.studies import StudyConfig, get_default_study_name
from syftr.tuner.qa_tuner import build_flow, eval_dataset

LLM_MAP = {
    "gpt-4o-mini": "gpt-4o-std",
    "anthropic-haiku-35": "anthropic-sonnet-35",
    "gemini-flash": "gemini-pro",
}


def replace_models(
    flow_params: T.Dict[str, T.Any],
    study_config: StudyConfig,
    new_study: optuna.Study,
    parent_number: int,
):
    old_flow_params = flow_params.copy()
    for param, value in flow_params.items():
        if (param.endswith("llm") or param.endswith("llm_name")) and value in LLM_MAP:
            logger.info(
                "Replacing %s with %s for %s in flow params",
                value,
                LLM_MAP[value],
                param,
            )
            flow_params[param] = LLM_MAP[value]

    logger.info("Building pareto flow from params: %s", flow_params)
    try:
        flow = build_flow(study_config=study_config, params=flow_params)
    except Exception:
        logger.exception("Failed to build flow from %s", flow_params)
        return
    logger.info("Evaluating updated flow and attaching it to the trial...")
    try:
        eval_results = eval_dataset(
            study_config, study_config.dataset, flow, study_config.evaluation.mode
        )
    except Exception:
        logger.exception("Failed to evaluate resulting flow %s", flow)
        return
    obj1 = eval_results[study_config.optimization.objective_1_name]
    obj2 = eval_results[study_config.optimization.objective_2_name]
    distributions = study_config.search_space.build_distributions(
        params=old_flow_params
    )
    logger.info("Saving the updated flow...")
    if "enforce_full_evaluation" in old_flow_params:
        old_flow_params.pop("enforce_full_evaluation")
    new_trial = optuna.create_trial(
        values=[obj1, obj2],
        params=old_flow_params,
        distributions=distributions,
        state=optuna.trial.TrialState.COMPLETE,
    )
    new_trial.set_user_attr("parent_number", parent_number)
    new_study.add_trial(new_trial)
    logger.info("Done replacing the models for %s", flow_params)


@ray.remote
def replace_models_remote(
    flow_params: T.Dict[str, T.Any],
    study_config: StudyConfig,
    new_study: optuna.Study,
    parent_number: int,
):
    """A wrapper for running model replacement on a Ray cluster."""
    return replace_models(flow_params, study_config, new_study, parent_number)


def run_pareto_flows_replace_models(study_path: str, remote: bool = False):
    """
    Reads Pareto flows, replaces models in them and re-evaluates resulting flow.
    """
    logger.info("Loading study config from file %s", study_path)
    study_config = StudyConfig.from_file(study_path)
    logger.info("Loading pareto flows for study %s", study_config.name)
    storage = cfg.database.get_optuna_storage()
    study = optuna.load_study(
        study_name=study_config.name,
        storage=storage,
    )
    df_pareto = get_pareto_df(study_config, success_rate=0.5)
    new_study_name = f"{study_config.name}_large_models"
    logger.info("Creating new study %s for updated trials...", new_study_name)
    po_study = optuna.create_study(
        study_name=new_study_name,
        directions=study.directions,
        storage=storage,
        load_if_exists=True,
    )
    logger.info("Running the replacement...")
    if remote:
        opts: T.List[T.Any] = []
        for _, row in df_pareto.iterrows():
            flow_params = json.loads(row["user_attrs_flow"])
            parent_number = row["number"]
            opts.append(
                replace_models_remote.remote(
                    flow_params, study_config, po_study, parent_number
                )
            )
        ray.get(opts)
    else:
        for _, row in df_pareto.iterrows():
            parent_number = row["number"]
            flow_params = json.loads(row["user_attrs_flow"])
            replace_models(flow_params, study_config, po_study, parent_number)
    logger.info("Done")


def run_opt(study_path: str, remote: bool):
    """Main workflow wrapper."""
    if remote:
        ray_init()
    run_pareto_flows_replace_models(study_path, remote=remote)


def main():
    """An argument parser."""
    parser = argparse.ArgumentParser()
    default = get_default_study_name()
    parser.add_argument(
        "--study-config",
        help=f"Path to study config yaml (default: {default})",
        default=default,
    )
    parser.add_argument(
        "--remote",
        help="Run the script on Ray",
        default=False,
        action="store_true",
    )
    args = parser.parse_args()
    run_opt(
        study_path=args.study_config,
        remote=args.remote,
    )


if __name__ == "__main__":
    main()
